"""mcms-benchs: A toolbox that provides a common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from numpy import pi as pi
import scipy.signal as sg
import string
from math import factorial


class Signal(np.ndarray):
    """
    A class that emulates a signal by behaving like a Numpy array for practical
    purposes but containing information of the signal such as the number of components
    and their instantaneous frequency (IF). This class is used in two ways:

    1. As a way of creating more complicated signals (i.e. with more components) and
    automatically stablish the number of components and IF by just defining them for
    the more simple, monocomponent signals. This way, when linearly combining two signals
    all the information regarding the combined components and their instantaneous
    frequencies is saved and is accessible from the final signal.

    2. As a way of wrap up information of the signal when passed to a method in the
    Benchmark class. The idea is that methods called by this latter class only receive
    a signal and parameters for the method. Therefore, by encapsulating the information
    of the signal in the Signal object, the method can use it while keep treating the
    received signal as a regular numpy array.

    A Signal class object has four attributes that differentiate it from a regular
    numpy array:
    1. comps: A list with each of the individual components combined to produce the
    signal.
    2. insft: A list with each of the individual components' instantenous frequency.
    The length of this list is the same as the comps list.
    3. ncomps: A numpy array indicating the number of components present in each time
    sample of the signal.
    4. total_comps: The total amount of components present within the duration of the
    signal. This is simply the length of comps.

    Methods
    -------

        def add_comp(self, new_comp, **kwargs)
        def add_instf(self, new_instf, **kwargs)


    """

    def __new__(
        subtype, array, instf=None, buffer=None, offset=0, strides=None, order=None
    ):

        dtype = array.dtype
        shape = array.shape
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to InfoArray.__array_finalize__
        obj = super().__new__(subtype, shape, dtype, buffer, offset, strides, order)
        # set the new 'info' attribute to the value passed

        if len(array) == 1:
            return array

        # Get the values of the ndarray
        obj[:] = array[:]

        # Add new attributes to the object
        obj._comps = [
            array.copy(),
        ]
        obj._ncomps = None
        obj._total_comps = None

        if instf is None:
            obj._instf = [
                np.zeros_like(array),
            ]
        else:
            obj._instf = [
                instf.copy(),
            ]

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(InfoArray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. InfoArray():
        #    obj is None
        #    (we're in the middle of the InfoArray.__new__
        #    constructor, and self.info will be set when we return to
        #    InfoArray.__new__)
        if obj is None:
            return
        # From view casting - e.g arr.view(InfoArray):
        #    obj is arr
        #    (type(obj) can be InfoArray)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is InfoArray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'info', because this
        # method sees all creation of default objects - with the
        # InfoArray.__new__ constructor, but also with
        # arr.view(InfoArray).
        self._comps = getattr(
            obj,
            "_comps",
            [
                obj,
            ],
        )
        self._instf = getattr(obj, "_instf", list())
        self._ncomps = getattr(obj, "_ncomps", None)
        self._total_comps = getattr(obj, "_total_comps", None)
        # self._instf = getattr(obj, '_ncomps', [obj, ])

        # [np.zeros_like(self._comps[0]),]

        # self._ncomps = getattr(obj,'_ncomps', None)
        # We do not need to return anything

    # def __init__(self, array=None, instf=None):
    #     self.comps = list()
    #     self.comps.append(array)

    #     if instf is None:
    #         self.instf = np.zeros_like(array)
    #     else:
    #         self.instf = instf

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        args = []

        for i, input_ in enumerate(inputs):
            # args.append(input_)
            if isinstance(input_, Signal):
                aux = input_.view(np.ndarray)
                if len(aux) == 1:
                    aux = aux[0]

                args.append(aux)
            else:
                args.append(input_)

        results = super().__array_ufunc__(ufunc, method, *tuple(args), **kwargs)

        if isinstance(results, np.ndarray):  # and ufunc.__name__ =='__add__':
            results = results.view(Signal)
            results._comps = []
            results._instf = []
            for ip in [a_signal for a_signal in inputs if isinstance(a_signal, Signal)]:
                for cp, instf in zip([*ip.comps], [*ip.instf]):
                    results.add_comp(cp, instf=instf)

        return results

    # Other functions from numpy.ndarray that we need.
    # TODO: Generalize these methods on one super() based function.
    def std(self, axis, dtype, out, ddof, **kwargs):
        return self.view(np.ndarray).std(
            axis=axis, dtype=dtype, out=out, ddof=ddof, **kwargs
        )

    def var(self, axis, dtype, out, ddof, **kwargs):
        return self.view(np.ndarray).var(
            axis=axis, dtype=dtype, out=out, ddof=ddof, **kwargs
        )

    @property
    def total_comps(self):
        if self._total_comps is None:
            self._total_comps = len(self._comps)
        return self._total_comps

    @total_comps.setter
    def total_comps(self, value):
        self._total_comps = value

    @property
    def ncomps(self):
        if self._ncomps is None:
            self.component_counter()
        return self._ncomps

    @ncomps.setter
    def ncomps(self, value):
        self._ncomps = value

    @property
    def comps(self):
        return self._comps

    @property
    def instf(self):
        return self._instf

    def add_comp(self, new_comp, **kwargs):
        """Add a new component, potentially with its instantaneous frequency. This
        latter can be added later.

        Args:
            new_comp (numpy.ndarray): New component to add.
            instf (numpy.ndarray): The instantaneous frequency corresponding to new_comp
        """

        self._comps.append(new_comp)
        if "instf" in kwargs.keys():
            self._instf.append(kwargs["instf"])

    def add_instf(self, new_instf, **kwargs):
        """Add a new instantaneous frequency.

        Args:
            new_instf (numpy.ndarray): A vector with the instantaneous frequency.
        """
        self._instf.append(new_instf)

    def component_counter(self):
        """This functions counts the number of components per time sample, and set the
        corresponding attribute self._ncomps.
        """
        N = len(self)
        cc = np.zeros((N,), dtype=int)
        th = 0.0
        for component in self._comps:
            for i in range(N):
                if np.sum(np.abs(component[i])) > th:
                    cc[i] += 1
        self._ncomps = cc

    def get_info(self):
        return {
            "ndarray": self.view(np.ndarray),
            "ncomps": self.ncomps,
            "total_comps": self.total_comps,
        }


class SignalBank:
    """
    Create a bank of signals. This class encapsulates the signal generation code,
    and returns different methods to generate signals as well as a dictionary of those
    methods that can be used later. Methods starting with "signal" generate
    monocomponent signals. Methods starting with "signal_mc" generate multicomponent
    signals.

    Both types of signals al generated with a length "N" passed as input parameter at
    the moment of instantiation. Signals are separated at least N^0.5 samples from the
    borders of the time-frequency plane in order to reduce border effects.

    Methods
    -------
        def check_frec_margins(self, instf):
            Check that the instantaneous frequency (if available) of a generated signal
            is withing certain margins to avoid aliasing and border effects.

        def generate_signal_dict(self):
            This function is used by the class constructor to generate a dictionary of
            signals. The keys of this dictionary are the name of the signals or
            "signal_id" that is used to indicate the benchmark which signals use to
            compare methods.

        def get_signal_id(self):
            Get the keys of the dictionary of signals generated by the function
            "generate_signal_dict()" when needed.

        def signal_linear_chirp(self, a=None, b=None, instfreq = False):
            Returns a linear chirp, the instantaneous frequency of which is a linear
            function with slope "a" and initial normalized frequency "b".

        def signal_mc_crossing_chirps(self):
            Returns a multi component signal with two chirps crossing, i.e. two chirps
            whose instantaneous frequency coincide in one point of the time frequency
            plane.

        def signal_mc_pure_tones(self, ncomps=5, a1=None, b1=None):
            Generates a multicomponent signal comprising several pure tones harmonically
            separated, i.e. tones are ordered from lower to higher frequency and each
            one has an instantaneous frequency that is an entire multiple of that of the
            previous tone.

        def signal_mc_multi_linear(self, ncomps=5):
            Generates a multicomponent signal with multiple linear chirps.

        def signal_tone_damped(self):
            Generates a damped tone whose normalized frequency is 0.25.

        def signal_tone_sharp_attack(self):
            Generates a damped tone that is modulated with a rectangular window.

        def signal_cos_chirp(self, omega=1.5, a1=1, f0=0.25, a2=0.125, checkinstf=True):
            Generates a cosenoidal chirp, the instantenous frequency of which is given
            by the formula: "f0 + a1*cos(2*pi*omega)", and the maximum amplitude of
            which is determined by "a2".

        def signal_mc_double_cos_chirp(self):
            Generates a multicomponent signal with two cosenoidal chirps.

        def signal_mc_cos_plus_tone(self):
            Generates a multicomponent signal comprised by two cosenoidal chirps and a
            single tone.

        def signal_mc_synthetic_mixture(self):
            Generates a multicomponent signal with different types of components.

        def signal_hermite_function(self, order = 18, t0 = 0.5, f0 = 0.25):
            Generates a round hermite function of a given order. The spectrogram of
            Hermite functions are given by an annular ridge in the time frequency plane,
            the center of which is given by (t0,f0).

        def signal_hermite_elipse(self, order = 30, t0 = 0.5, f0 = 0.25):
            Generates a non-round Hermite function of a given order. The spectrogram of
            Hermite functions are given by an elipsoidal ridge in the time frequency
            plane the center of which is given by (t0,f0).

        def signal_mc_triple_impulse(self, Nimpulses = 3):
            Generates three equispaced impulses in time.

        def signal_mc_impulses(self, Nimpulses = 7):
            Generates equispaced impulses in time.

        def signal_exp_chirp(self, finit=None, fend=None, exponent=2, r_instf=False):
            Generates an exponential chirp.

        def signal_mc_exp_chirps(self):
            Generates a multicomponent signal comprising three exponential chirps.

        def signal_mc_multi_cos(self):
            Generates a multicomponent signal comprising three cosenoidal chirps with
            different frequency modulation parameters.

        def signal_mc_synthetic_mixture_2(self):
            Generates a multicomponent signal with different types of components.

        def signal_mc_on_off_tones(self):
            Generates a multicomponent signal comprising components that "born" and
            "die" at different times.

        def signal_mc_synthetic_mixture_3(self):
            Generates a multicomponent signal with different types of components.

        def get_all_signals(self):
            Returns an array of shape [K,N] where K is the number of signals generated
            by this signal bank, and N is the length of the signals.
    """

    def __init__(self, N=2**8, Nsub=None, return_signal=False):
        """Builds a dictionary of functions that return multiple signals.
        Args:
            N (int, optional): Length of the signals. Defaults to 2**8.
            Nsub (int, optional): Generates signals of length Nsub,
            then zero pads the vector to reach length N. Defaults to 2**8.
            return_signal (bool, optional): If True, functions will return a Signal
            object, that encapsulates more information of the signal such as the number
            of components, each individual component and their instantaneous frequency.
        """
        self.return_signal = return_signal

        self.N = N
        if Nsub is None:
            self.tmin = int(np.sqrt(N))
            self.tmax = self.N - self.tmin
            self.Nsub = self.tmax - self.tmin
        else:
            self.Nsub = Nsub
            self.tmin = (self.N - self.Nsub) // 2
            self.tmax = self.tmin + Nsub

        # self.fmin = 1.0*np.sqrt(N)/N
        # self.fmax = 0.5-self.fmin
        # self.fmid = (self.fmax-self.fmin)/2 + self.fmin

        self.fmin = 0.07
        self.fmax = 0.5 - self.fmin
        self.fmid = 0.25

        # print(self.fmin)
        # print(self.fmax)

        self.generate_signal_dict()

    # TODO
    def check_inst_freq(self, instf):
        """Check that the instantaneous frequency (if available) of a generated signal
        is withing certain margins to avoid aliasing and border effects.

        Args:
            instf (numpy.ndarray): Instantaneous frequency of a signal.
        """

        # if np.all(instf<=self.fmax):
        #     print('Warning: instf>fmax')
        # if np.all(instf>=self.fmin):
        #     print('Warning: instf<fmin')
        return True

    def generate_signal_dict(self):
        """This function is used by the class constructor to generate a dictionary of
        signals. The keys of this dictionary are the name of the signals or "signal_id"
        that is used to indicate the benchmark which signals use to compare methods.

        Returns:
            dict: Dictionary of functions that returns a signal when called.
        """

        campos = dir(self)
        fun_names = [fun_name for fun_name in campos if fun_name.startswith("signal_")]
        signal_ids = [
            string.capwords(fun_name[7::], sep="_").replace("_", "")
            for fun_name in fun_names
        ]

        self.signalDict = dict()
        for i, signal_id in enumerate(signal_ids):
            try:
                self.signalDict[signal_id] = getattr(self, fun_names[i])()

            except BaseException as err:
                self.signalDict[signal_id] = None

        return self.signalDict

    def get_signal_id(self):
        """Get the keys of the dictionary of signals generated by the function
        "generate_signal_dict()" when needed.

        Returns:
            tuple: Tuple with the keys of a dictionary of signals.
        """

        return self.SignalDict.keys()

    def get_all_signals(self):
        """Returns an array of shape [K,N] where K is the number of signals generated
        by this signal bank, and N is the length of the signals.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        signals = np.zeros((len(self.signalDict), self.N))
        for k, key in enumerate(self.signalDict):
            signals[k] = self.signalDict[key]()
        return signals

    # Monocomponent signals --------------------------------------------------------

    def _signal_linear_chirp(self, a=None, b=None, phi=0, instfreq=False):
        """Returns a linear chirp, the instantaneous frequency of which is a linear
        function with slope "a" and initial normalized frequency "b".


        Args:
            a (int, optional): Slope of the instantaneous frequency. Defaults to None.
            b (int, optional): Initial instantaneous frequency. Defaults to None.
            instfreq (bool, optional): When True, returns a vector with the
            instantaneous frequency. Defaults to False.

        Returns:
            list or ndarray: If input parameter "instfreq" is True, returns the a list
            of ndarray type objects with the signal and its instantaneous frequency.
        """

        N = self.N

        if a is None:
            a = self.fmax - self.fmin
        if b is None:
            b = self.fmin

        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub

        tsub = np.arange(Nsub)
        instf = b + a * tsub / Nsub

        if not instfreq:
            self.check_inst_freq(instf)

        phase = np.cumsum(instf)

        x = np.cos(2 * pi * phase + phi) * sg.windows.tukey(Nsub, 0.25)
        signal = np.zeros((N,))
        signal[tmin:tmax] = x

        emb_instf = np.zeros_like(signal)
        emb_instf[tmin:tmax] = instf

        # Cast to Signal class.
        signal = Signal(array=signal, instf=emb_instf)

        if instfreq:
            return signal, instf, tmin, tmax
        else:
            return signal

    def _signal_tone_damped(self):
        """Generates a damped tone whose normalized frequency is 0.25.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        eps = 1e-6
        t = np.arange(N) + eps
        c = 1 / N / 10
        prec = 1e-1  # Precision at sample N for the envelope.
        alfa = -np.log(prec * N / ((N - c) ** 2)) / N
        e = np.exp(-alfa * t) * ((t - c) ** 2 / t)
        e[0] = 0
        chirp = self._signal_linear_chirp(a=0, b=0.25)
        return e * chirp

    def _signal_tone_sharp_attack(self):
        """Generates a damped tone that is modulated with a rectangular window.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        dumpcos = self.signal_tone_damped()
        indmax = np.argmax(dumpcos)
        dumpcos[0:indmax] = 0
        return dumpcos

    def _signal_exp_chirp(self, finit=None, fend=None, exponent=2, r_instf=False):
        """Generates an exponential chirp.

        Args:
            finit (float, optional): Initial normalized frequency. Defaults to None.
            fend (float, optional): End normalized frequency. Defaults to None.
            exponent (int, optional): Exponent. Defaults to 2.
            r_instf (bool, optional): When True returns the instantaneous frequency
            along with the signal. Defaults to False.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub
        tsub = np.arange(Nsub) / Nsub

        if finit is None:
            finit = 1.5 * self.fmin

        if fend is None:
            fend = self.fmax

        instf = finit * np.exp(np.log(fend / finit) * tsub**exponent)

        if not r_instf:
            self.check_inst_freq(instf)

        phase = np.cumsum(instf)
        x = np.cos(2 * pi * phase)
        signal = np.zeros((N,))
        signal[tmin:tmax] = x * sg.windows.tukey(Nsub, 0.25)

        emb_instf = np.zeros_like(signal)
        emb_instf[tmin:tmax] = instf

        # Cast to Signal class.
        signal = Signal(array=signal, instf=emb_instf)

        if r_instf:
            return signal, instf, tmin, tmax
        else:
            return signal

    def _signal_cos_chirp(self, omega=1.2, a1=0.5, f0=0.25, a2=0.125, checkinstf=True):
        """Generates a cosenoidal chirp, the instantenous frequency of which is given by
        the formula: "f0 + a1*cos(2*pi*omega)", and the maximum amplitude of which is
        determined by "a2".

        Args:
            omega (float, optional): Frequency of the instantaneous frequency.
            Defaults to 1.5.
            a1 (int, optional): Amplitude of the frequency modulation Defaults to 1.
            f0 (float, optional): Central frequency. Defaults to 0.25.
            a2 (float, optional): Amplitude of the signal. Defaults to 0.125.
            checkinstf (bool, optional): If True checks that dhe instantaneous frequency
            of the signal is within the limits. Defaults to True.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub
        tsub = np.arange(Nsub)
        instf = f0 + a2 * np.cos(2 * pi * omega * tsub / Nsub - pi * omega)

        if checkinstf:
            self.check_inst_freq(instf)

        phase = np.cumsum(instf)
        x = a1 * np.cos(2 * pi * phase) * sg.windows.tukey(Nsub, 0.25)
        signal = np.zeros((N,))
        signal[tmin:tmax] = x

        emb_instf = np.zeros_like(signal)
        emb_instf[tmin:tmax] = instf

        # Cast to Signal class.
        signal = Signal(array=signal, instf=emb_instf)

        return signal

    # Output monocomponent signals -------------------------------------------------
    def signal_linear_chirp(self, a=None, b=None, instfreq=False):
        """Returns a linear chirp, the instantaneous frequency of which is a linear
        function with slope "a" and initial normalized frequency "b".


        Args:
            a (int, optional): Slope of the instantaneous frequency. Defaults to None.
            b (int, optional): Initial instantaneous frequency. Defaults to None.
            instfreq (bool, optional): When True, returns a vector with the
            instantaneous frequency. Defaults to False.

        Returns:
            list or ndarray: If input parameter "instfreq" is True, returns the a list
            of ndarray type objects with the signal and its instantaneous frequency.
        """

        signal = self._signal_linear_chirp(a=a, b=b, instfreq=instfreq)
        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_tone_damped(self):
        """Generates a damped tone whose normalized frequency is 0.25.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        eps = 1e-6
        t = np.arange(N) + eps
        c = 1 / N / 10
        prec = 1e-1  # Precision at sample N for the envelope.
        alfa = -np.log(prec * N / ((N - c) ** 2)) / N
        e = np.exp(-alfa * t) * ((t - c) ** 2 / t)
        e[0] = 0
        chirp = self._signal_linear_chirp(a=0, b=0.25)
        signal = e * chirp

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    def signal_tone_sharp_attack(self):
        """Generates a damped tone that is modulated with a rectangular window.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        signal = self.signal_tone_damped()
        indmax = np.argmax(signal)
        signal[0:indmax] = 0

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    def signal_exp_chirp(self, finit=None, fend=None, exponent=2, r_instf=False):
        """Generates an exponential chirp.

        Args:
            finit (float, optional): Initial normalized frequency. Defaults to None.
            fend (float, optional): End normalized frequency. Defaults to None.
            exponent (int, optional): Exponent. Defaults to 2.
            r_instf (bool, optional): When True returns the instantaneous frequency
            along with the signal. Defaults to False.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """
        signal = self._signal_exp_chirp(
            finit=finit, fend=fend, exponent=exponent, r_instf=r_instf
        )
        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_cos_chirp(self, omega=1.2, a1=0.5, f0=0.25, a2=0.125, checkinstf=True):
        """Generates a cosenoidal chirp, the instantenous frequency of which is given by
        the formula: "f0 + a1*cos(2*pi*omega)", and the maximum amplitude of which is
        determined by "a2".

        Args:
            omega (float, optional): Frequency of the instantaneous frequency.
            Defaults to 1.5.
            a1 (int, optional): Amplitude of the frequency modulation Defaults to 1.
            f0 (float, optional): Central frequency. Defaults to 0.25.
            a2 (float, optional): Amplitude of the signal. Defaults to 0.125.
            checkinstf (bool, optional): If True checks that dhe instantaneous frequency
            of the signal is within the limits. Defaults to True.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """
        signal = self._signal_cos_chirp(
            omega=omega, a1=a1, f0=f0, a2=a2, checkinstf=checkinstf
        )
        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    # Multicomponent signals here --------------------------------------------------

    def signal_mc_parallel_chirps(self):
        comp1 = self._signal_linear_chirp(a=0.1, b=0.15)
        comp2 = self._signal_linear_chirp(a=0.1, b=0.25)
        signal = comp1 + comp2

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    def signal_mc_parallel_chirps_unbalanced(self):
        comp1 = self._signal_linear_chirp(a=0.1, b=0.15, instfreq=False)
        comp2 = self._signal_linear_chirp(a=0.1, b=0.25, instfreq=False)
        signal = comp1 + 0.2 * comp2

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    def signal_mc_on_off_2(self):
        chirp1 = self._signal_linear_chirp(a=0.1, b=0.10, instfreq=False)
        chirp2a = self._signal_linear_chirp(a=0.1, b=0.20, instfreq=False)
        chirp2b = self._signal_linear_chirp(a=0.1, b=0.20, instfreq=False)
        chirp3 = self._signal_linear_chirp(a=0.1, b=0.30, instfreq=False)

        Nsub = self.N
        N3 = Nsub // 3
        N4 = Nsub // 4
        N7 = Nsub // 7
        N9 = Nsub // 9

        chirp1[0 : 2 * N7] = 0
        chirp1.comps[0][0 : 2 * N7] = 0
        chirp1.instf[0][0 : 2 * N7] = 0

        chirp1[5 * N7 : -1] = 0
        chirp1.comps[0][5 * N7 : -1] = 0
        chirp1.instf[0][5 * N7 : -1] = 0

        chirp1[2 * N7 : 5 * N7] = chirp1[2 * N7 : 5 * N7] * sg.windows.tukey(
            3 * N7, 0.25
        )
        chirp1.comps[0][2 * N7 : 5 * N7] = chirp1[2 * N7 : 5 * N7] * sg.windows.tukey(
            3 * N7, 0.25
        )

        chirp2a[0:N9] = 0
        chirp2a.comps[0][0:N9] = 0
        chirp2a.instf[0][0:N9] = 0
        chirp2a[8 * N9 : -1] = 0
        chirp2a.comps[0][8 * N9 : -1] = 0
        chirp2a.instf[0][8 * N9 : -1] = 0
        chirp2a[4 * N9 : 5 * N9] = 0
        chirp2a.comps[0][4 * N9 : 5 * N9] = 0
        chirp2a.instf[0][4 * N9 : 5 * N9] = 0
        chirp2a[N9 : 4 * N9] = chirp2a[N9 : 4 * N9] * sg.windows.tukey(3 * N9, 0.25)
        chirp2a.comps[0][N9 : 4 * N9] = chirp2a[N9 : 4 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2a._instf[0][5 * N9 : 8 * N9] = 0
        chirp2a.comps[0][5 * N9 : 8 * N9] = 0
        chirp2a[5 * N9 : 8 * N9] = 0

        chirp2b[0:N9] = 0
        chirp2b.comps[0][0:N9] = 0
        chirp2b.instf[0][0:N9] = 0
        chirp2b[8 * N9 : -1] = 0
        chirp2b.comps[0][8 * N9 : -1] = 0
        chirp2b.instf[0][8 * N9 : -1] = 0
        chirp2b[4 * N9 : 5 * N9] = 0
        chirp2b.comps[0][4 * N9 : 5 * N9] = 0
        chirp2b.instf[0][4 * N9 : 5 * N9] = 0
        chirp2b[5 * N9 : 8 * N9] = chirp2b[5 * N9 : 8 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2b.comps[0][5 * N9 : 8 * N9] = chirp2b[5 * N9 : 8 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2b._instf[0][N9 : 4 * N9] = 0
        chirp2b.comps[0][N9 : 4 * N9] = 0
        chirp2b[N9 : 4 * N9] = 0

        signal = chirp1 + chirp2b + chirp2a + chirp3

        if not self.return_signal:
            signal = signal.view(np.ndarray)

        return signal

    def signal_mc_crossing_chirps(self):
        """Returns a multi component signal with two chirps crossing, i.e. two chirps
        whose instantaneous frequency coincide in one point of the time frequency plane.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N

        a = self.fmax - self.fmin
        b = self.fmin

        chirp1 = 0.8 * self._signal_linear_chirp(a=-0.31, b=0.41, phi=0)
        chirp2 = self._signal_linear_chirp(a=0.35, b=0.1, phi=0)

        signal = chirp1 + chirp2

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    # def signal_mc_crossing_chirps_2(self):
    #     """Returns a multi component signal with two chirps crossing, i.e. two chirps
    #     whose instantaneous frequency coincide in one point of the time frequency plane.

    #     Returns:
    #         numpy.ndarray: Returns a numpy array with the signal.
    #     """

    #     N = self.N

    #     a = self.fmax-self.fmin
    #     b = self.fmin

    #     chirp1 = self._signal_linear_chirp(a = -a, b = 0.5 - b)
    #     chirp2 = self._signal_linear_chirp(a = a, b = b)

    #     tmin = self.tmin
    #     tmax = N-tmin
    #     Nsub = tmax-tmin
    #     tsub = np.arange(Nsub)
    #     fmax = self.fmax

    #     signal = np.zeros((N,))
    #     instf0 = np.zeros_like(signal)

    #     omega = 1.5
    #     f0 = 0.5 - b - a*tsub/Nsub
    #     if0 = f0+ 0.02 + 0.02*np.cos(2*pi*omega*tsub/Nsub - pi*omega)
    #     if0 = if0[np.where(if0<fmax)]
    #     self.check_inst_freq(if0)
    #     phase0 = np.cumsum(if0)
    #     x0 = np.zeros_like(signal)
    #     x0[tmin:tmin+len(phase0)] = np.cos(2*pi*phase0)*sg.windows.tukey(len(phase0),0.25)
    #     instf0[tmin:tmin+len(if0)] = if0
    #     chirp1 = Signal(x0, instf=instf0)

    #     signal = chirp1 + chirp2

    #     if not self.return_signal:
    #         return signal.view(np.ndarray)

    #     return signal

    def signal_mc_pure_tones(self, ncomps=5, a1=None, b1=None, c0=0.0):
        """Generates a multicomponent signal comprising several pure tones harmonically
        separated, i.e. tones are ordered from lower to higher frequency and each one
        has an instantaneous frequency that is an entire multiple of that of the
        previous tone.

        Args:
            ncomps (int, optional): Number of components. Defaults to 5.
            a1 (float, optional): Slope of chirps. Defaults to 0.
            b1 (float, optional): Frequency of the first tone. Defaults to None.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        k = 1
        aux = np.zeros((N,))
        max_freq = self.fmax

        if a1 is None:
            a1 = 0
        if b1 is None:
            b1 = self.fmid / 2
            if b1 < self.fmin:
                b1 = self.fmin

        signal = self._signal_linear_chirp(a=a1, b=b1 + c0)

        for i in range(1, ncomps):
            chirp = self._signal_linear_chirp(a=a1 * (i + 1), b=b1 * (i + 1) + c0)
            if np.max(chirp.instf[0]) >= max_freq:
                break
            signal = signal + chirp

        # for i in range(ncomps):
        #     chirp, instf, tmin, _ = self.signal_linear_chirp(a = a1*(i+1),
        #                                                      b = b1*(i+1),
        #                                                      instfreq=True)
        #     if instf[0] >= max_freq:
        #         break

        #     idx = np.where(instf < max_freq)[0] + tmin -1
        #     tukwin = sg.windows.tukey(idx.shape[0],0.5)
        #     chirp[idx] = chirp[idx]*tukwin
        #     idx = np.where(instf >= max_freq)[0] + tmin -1
        #     chirp[idx] = tukwin[-1]

        #     aux += chirp

        if not self.return_signal:
            return signal.view(np.ndarray)

        return signal

    def signal_mc_multi_linear(self, ncomps=5):
        """Generates a multicomponent signal with multiple linear chirps.

        Args:
            ncomps (int, optional): Number of components. Defaults to 5.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        a1 = self.fmin / 2
        b1 = self.fmin
        signal = self.signal_mc_pure_tones(ncomps=ncomps, a1=a1, b1=b1)

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_multi_linear_2(self, ncomps=5):
        """Generates a multicomponent signal with multiple linear chirps.

        Args:
            ncomps (int, optional): Number of components. Defaults to 5.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        a1 = self.fmin / 2
        b1 = self.fmin / 3
        signal = self.signal_mc_pure_tones(ncomps=ncomps, a1=a1, b1=b1, c0=self.fmin)

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_double_cos_chirp(self):
        """Generates a multicomponent signal with two cosenoidal chirps.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        t = np.arange(N) / N
        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub
        tsub = np.arange(Nsub)
        omega1 = 1.5
        omega2 = 1.8

        chirp1 = self._signal_cos_chirp(
            omega=omega1, a1=1.0, f0=self.fmax - 0.075, a2=0.05
        )

        chirp2 = self._signal_cos_chirp(
            omega=omega2, a1=1.0, f0=self.fmin + 0.075, a2=0.06
        )

        # instf1 = self.fmax-0.075 + 0.05*np.cos(2*pi*omega1*tsub/Nsub - pi*omega1)
        # self.check_inst_freq(instf1)
        # instf2 = self.fmin+0.075 + 0.06*np.cos(2*pi*omega2*tsub/Nsub - pi*omega2)
        # self.check_inst_freq(instf2)

        # phase1 = np.cumsum(instf1)
        # phase2 = np.cumsum(instf2)
        # x = np.cos(2*pi*phase1) + np.cos(2*pi*phase2)
        # x = x*sg.windows.tukey(Nsub,0.25)
        # signal = np.zeros((N,))
        # signal[tmin:tmax] = x

        signal = chirp1 + chirp2

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_cos_plus_tone(self):
        """Generates a multicomponent signal comprised by two cosenoidal chirps and a
        single tone.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        omega1 = 1.5
        omega2 = 1.8

        chirp1 = self._signal_cos_chirp(
            omega=omega1, a1=1.0, f0=self.fmax - 0.05, a2=0.04
        )

        chirp2 = self._signal_cos_chirp(omega=omega2, a1=1.0, f0=self.fmid, a2=0.04)

        chirp3 = self._signal_linear_chirp(a=0, b=1.8 * self.fmin)

        # instf1 = self.fmax-0.05 + 0.04*np.cos(2*pi*omega1*tsub/Nsub - pi*omega1)
        # self.check_inst_freq(instf1)
        # instf2 = self.fmid + 0.04*np.cos(2*pi*omega2*tsub/Nsub - pi*omega2)
        # self.check_inst_freq(instf2)
        # instf3 = 1.8*self.fmin * np.ones((Nsub,))
        # self.check_inst_freq(instf3)

        # phase1 = np.cumsum(instf1)
        # phase2 = np.cumsum(instf2)
        # phase3 = np.cumsum(instf3)
        # x = np.cos(2*pi*phase1) + np.cos(2*pi*phase2) + np.cos(2*pi*phase3)
        # x = x*sg.windows.tukey(Nsub,0.25)
        # signal = np.zeros((N,))
        # signal[tmin:tmax] = x

        signal = chirp1 + chirp2 + chirp3

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_multi_cos(self):
        """Generates a multicomponent signal comprising three cosenoidal chirps with
        different frequency modulation parameters.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        x1 = self._signal_cos_chirp(omega=8, a1=1, f0=self.fmin + 0.04, a2=0.03)
        x2 = self._signal_cos_chirp(omega=6, a1=1, f0=self.fmid, a2=0.02)
        x3 = self._signal_cos_chirp(omega=4, a1=1, f0=self.fmax - 0.03, a2=0.02)
        # x4 = self.signal_cos_chirp(omega = 10, a1=1, f0=0.42, a2=0.05)
        signal = x1 + x2 + x3

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_multi_cos_2(self):
        """Generates a multicomponent signal comprising three cosenoidal chirps with
        different frequency modulation parameters.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        x1 = self._signal_cos_chirp(omega=5, a1=1.5, f0=self.fmin + 0.04, a2=0.03)
        x2 = self._signal_cos_chirp(omega=5, a1=1.2, f0=self.fmid, a2=0.02)
        x3 = self._signal_cos_chirp(omega=5, a1=1, f0=self.fmax - 0.03, a2=0.02)
        # x4 = self.signal_cos_chirp(omega = 10, a1=1, f0=0.42, a2=0.05)
        signal = x1 + x2 + x3

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_synthetic_mixture(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        t = np.arange(N) / N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        fmax = self.fmax

        signal = np.zeros((N,))
        instf0 = np.zeros_like(signal)
        instf1 = np.zeros_like(signal)
        instf2 = np.zeros_like(signal)

        omega = 7
        f0 = self.fmin + 0.07 * tsub / Nsub
        f1 = 1.2 * self.fmin + 0.25 * tsub / Nsub
        f2 = 1.3 * self.fmin + 0.53 * tsub / Nsub

        if0 = f0 + 0.02 + 0.02 * np.cos(2 * pi * omega * tsub / Nsub - pi * omega)
        if1 = f1 + 0.02 + 0.02 * np.cos(2 * pi * omega * tsub / Nsub - pi * omega)
        if2 = f2 + 0.02 + 0.02 * np.cos(2 * pi * omega * tsub / Nsub - pi * omega)
        if0 = if0[np.where(if0 < fmax)]
        if1 = if1[np.where(if1 < fmax)]
        if2 = if2[np.where(if2 < fmax)]

        self.check_inst_freq(if0)
        self.check_inst_freq(if1)
        self.check_inst_freq(if2)

        phase0 = np.cumsum(if0)
        phase1 = np.cumsum(if1)
        phase2 = np.cumsum(if2)

        x0 = np.zeros_like(signal)
        x1 = np.zeros_like(signal)
        x2 = np.zeros_like(signal)

        x0[tmin : tmin + len(phase0)] = np.cos(2 * pi * phase0) * sg.windows.tukey(
            len(phase0), 0.25
        )
        x1[tmin : tmin + len(phase1)] = np.cos(2 * pi * phase1) * sg.windows.tukey(
            len(phase1), 0.25
        )
        x2[tmin : tmin + len(phase2)] = np.cos(2 * pi * phase2) * sg.windows.tukey(
            len(phase2), 0.25
        )

        instf0[tmin : tmin + len(if0)] = if0
        instf1[tmin : tmin + len(if1)] = if1
        instf2[tmin : tmin + len(if2)] = if2

        chirp0 = Signal(x0, instf=instf0)
        chirp1 = Signal(x1, instf=instf1)
        chirp2 = Signal(x2, instf=instf2)

        # signal = x0+x1+x2

        signal = chirp0 + chirp1 + chirp2

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_synthetic_mixture_2(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        def rect_window(N, ti, te):
            rw = np.zeros((N,))
            rw[ti:te] = 1
            return rw

        N = self.N
        t = np.arange(N)
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        fmax = self.fmax
        fmin = self.fmin

        tt = Nsub // 2
        t_init = tmin

        f_init = (0.25 - fmin) / 2 + fmin
        f_end = fmax
        m = (f_end - f_init) / tt

        instf1 = (m * (t - t_init) + f_init) * rect_window(N, t_init, t_init + tt)
        phase1 = np.cumsum(instf1)
        x1 = np.cos(2 * pi * phase1) * rect_window(N, t_init, t_init + tt)
        x1[t_init : t_init + tt] *= sg.windows.tukey(tt, 0.4)

        c = 1 / tt / 10
        prec = 1e-1  # Precision at sample N for the envelope.
        alfa = -np.log(prec * tt / ((tt - c) ** 2)) / tt
        e = np.exp(-alfa * np.arange(tt)) * (
            (np.arange(tt) - c) ** 2 / (np.arange(tt) + 1e-15)
        )
        e[0] = 0
        e /= np.max(e)

        t_init += tt // 2
        instf2 = (m * (t - t_init) + f_init) * rect_window(N, t_init, t_init + tt)
        phase2 = np.cumsum(instf2)
        x2 = np.cos(2 * pi * phase2) * rect_window(N, t_init, t_init + tt)
        x2[t_init : t_init + tt] *= sg.windows.tukey(tt, 0.25) * e

        t_init += tt // 2
        instf3 = (m * (t - t_init) + f_init) * rect_window(N, t_init, t_init + tt)
        phase3 = np.cumsum(instf3)
        x3 = np.cos(2 * pi * phase3) * rect_window(N, t_init, t_init + tt)
        x3[t_init : t_init + tt] *= sg.windows.tukey(tt, 0.25) * e[-1::-1]

        x4 = np.cos(2 * pi * fmin * np.ones((N,)) * t) * rect_window(N, tmin, tmax)
        x4[tmin:tmax] *= sg.windows.tukey(Nsub, 0.75)
        instf4 = fmin * np.ones((N,))

        signal = (
            Signal(x1, instf=instf1)
            + Signal(x2, instf=instf2)
            + Signal(x3, instf=instf3)
            + Signal(x4, instf=instf4)
        )

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_synthetic_mixture_5(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub
        tmid = Nsub // 2
        # tmid = tmid +(tmax-tmid)//5
        tsub = np.arange(Nsub)
        signal = np.zeros((N,))

        if1 = 1 * self.fmin + 4.0 * (tsub / Nsub - 0.05) ** 2
        if1 = if1[np.where(if1 < self.fmax)]
        if1 = if1[np.where(self.fmin < if1)]
        phase1 = np.cumsum(if1)
        x = np.cos(2 * pi * phase1) * sg.windows.tukey(len(phase1), 0.5)
        chirp1 = np.zeros_like(signal)
        chirp1[tmin : tmin + len(x)] = x
        instf1 = np.zeros_like(signal)
        instf1[tmin : tmin + len(x)] = if1

        N = self.N
        tmin = 12 * self.tmin
        tmax = N - self.tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        fmax = self.fmax

        signal = np.zeros((N,))
        instf2 = np.zeros_like(signal)
        omega = 4
        f2 = 1.2 * self.fmin + 0.1 * tsub / Nsub
        if2 = f2 + 0.035 + 0.02 * np.cos(2 * pi * omega * tsub / Nsub - pi * omega)
        if2 = if2[np.where(if2 < fmax)]
        phase2 = np.cumsum(if2)
        x2 = np.zeros_like(signal)
        x2[tmin : tmin + len(phase2)] = np.cos(2 * pi * phase2) * sg.windows.tukey(
            len(phase2), 0.25
        )
        instf2[tmin : tmin + len(if2)] = if2

        tmin = 12 * self.tmin
        tmax = N - self.tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        instf3 = np.zeros_like(signal)
        omega = 5
        f3 = 1.2 * self.fmin + 0.2 * tsub / Nsub
        if3 = f3 + 0.17  # + 0.02*np.cos(2*pi*omega*tsub/Nsub - pi*omega)
        if3 = if3[np.where(if3 < fmax)]
        phase3 = np.cumsum(if3)
        x3 = np.zeros_like(signal)
        x3[tmin : tmin + len(phase3)] = np.cos(2 * pi * phase3) * sg.windows.tukey(
            len(phase3), 0.25
        )
        instf3[tmin : tmin + len(if3)] = if3

        signal = (
            Signal(chirp1, instf=instf1)
            + Signal(x2, instf=instf2)
            + Signal(x3, instf=instf3)
            # + Signal(chirp4, instf=None)
        )

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_synthetic_mixture_3(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = self.tmax
        Nsub = self.Nsub
        tmid = Nsub // 2
        # tmid = tmid +(tmax-tmid)//5
        tsub = np.arange(Nsub)
        signal = np.zeros((N,))

        if1 = 1 * self.fmin + 1 * (tsub / Nsub - 0.05) ** 2
        if1 = if1[np.where(if1 < self.fmax)]
        if1 = if1[np.where(self.fmin < if1)]
        phase1 = np.cumsum(if1)
        x = np.cos(2 * pi * phase1) * sg.windows.tukey(len(phase1), 0.5)
        chirp1 = np.zeros_like(signal)
        chirp1[tmin : tmin + len(x)] = x
        instf1 = np.zeros_like(signal)
        instf1[tmin : tmin + len(x)] = if1

        if2 = np.ones((Nsub,)) * (self.fmid - self.fmin) / 2
        if2 = if2[tmid::]
        phase2 = np.cumsum(if2)
        x2 = np.cos(2 * pi * phase2) * sg.windows.tukey(len(phase2), 0.1)
        chirp2 = np.zeros_like(signal)
        chirp2[tmid : tmid + len(x2)] = x2
        instf2 = np.zeros_like(signal)
        instf2[tmid : tmid + len(x2)] = if2

        instf3 = np.ones((N,)) * (2 * (self.fmid - self.fmin) / 3 + self.fmid)
        phase3 = np.cumsum(instf3)
        tloc = 5 * N // 7
        chirp3 = (
            3
            * np.cos(2 * pi * phase3)
            * np.exp(-np.pi * (np.arange(N) - tloc) ** 2 / (N / 8))
        )

        instf4 = np.ones((N,)) * ((self.fmid - self.fmin) / 3 + self.fmid)
        phase4 = np.cumsum(instf4)
        tloc = 6 * N // 7
        chirp4 = (
            5
            * np.cos(2 * pi * phase4)
            * np.exp(-np.pi * (np.arange(N) - tloc) ** 2 / (N / 8))
        )

        signal = (
            Signal(chirp1, instf=instf1)
            + Signal(chirp2, instf=instf2)
            + Signal(chirp3, instf=None)
            + Signal(chirp4, instf=None)
        )

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_synthetic_mixture_4(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        def rect_window(N, ti, te):
            rw = np.zeros((N,))
            rw[ti:te] = 1
            return rw

        N = self.N
        t = np.arange(N)
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        fmax = self.fmax
        fmin = self.fmin

        tt = int(Nsub // 2.5)
        t_init = 4 * tmin

        f_init = (0.25 - fmin) / 2 + fmin
        f_end = fmax
        m = (f_end - f_init) / tt

        c = 1 / tt / 10
        prec = 1e-1  # Precision at sample N for the envelope.
        alfa = -np.log(prec * tt / ((tt - c) ** 2)) / tt
        e = np.exp(-alfa * np.arange(tt)) * (
            (np.arange(tt) - c) ** 2 / (np.arange(tt) + 1e-15)
        )
        e[0] = 0
        e /= np.max(e)

        signal = np.zeros(
            N,
        )
        for i in range(1, 6):
            instf1 = np.zeros(
                N,
            )
            instf1[t_init : t_init + tt] = e * fmin * 0.9**i + i * fmin + 0.05
            phase1 = np.cumsum(instf1)
            x1 = (
                0.9 ** (i - 1)
                * np.cos(2 * pi * phase1)
                * rect_window(N, t_init, t_init + tt)
            )
            x1[t_init : t_init + tt] *= sg.windows.tukey(tt, 0.25)
            signal = signal + Signal(x1, instf=instf1)

        t_init += int(1.01 * tt)

        for i in range(1, 6):
            instf1 = np.zeros(
                N,
            )
            instf1[t_init : t_init + tt] = e * fmin * 0.9 + i * fmin + 0.03
            phase1 = np.cumsum(instf1)
            x1 = (
                0.8 ** (i - 1)
                * np.cos(2 * pi * phase1)
                * rect_window(N, t_init, t_init + tt)
            )
            x1[t_init : t_init + tt] *= sg.windows.tukey(tt, 0.5)
            signal = signal + Signal(x1, instf=instf1)

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_on_off_tones(self):
        """Generates a multicomponent signal comprising components that "born" and "die"
        at different times.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        chirp1 = self._signal_linear_chirp(a=0, b=0.10, instfreq=False)
        chirp2a = self._signal_linear_chirp(a=0, b=0.20, instfreq=False)
        chirp2b = self._signal_linear_chirp(a=0, b=0.20, instfreq=False)
        chirp3 = self._signal_linear_chirp(a=0, b=0.30, instfreq=False)

        Nsub = self.N
        N3 = Nsub // 3
        N4 = Nsub // 4
        N7 = Nsub // 7
        N9 = Nsub // 9

        chirp1[0 : 2 * N7] = 0
        chirp1.comps[0][0 : 2 * N7] = 0
        chirp1.instf[0][0 : 2 * N7] = 0

        chirp1[5 * N7 : -1] = 0
        chirp1.comps[0][5 * N7 : -1] = 0
        chirp1.instf[0][5 * N7 : -1] = 0

        chirp1[2 * N7 : 5 * N7] = chirp1[2 * N7 : 5 * N7] * sg.windows.tukey(
            3 * N7, 0.25
        )
        chirp1.comps[0][2 * N7 : 5 * N7] = chirp1[2 * N7 : 5 * N7] * sg.windows.tukey(
            3 * N7, 0.25
        )

        chirp2a[0:N9] = 0
        chirp2a.comps[0][0:N9] = 0
        chirp2a.instf[0][0:N9] = 0
        chirp2a[8 * N9 : -1] = 0
        chirp2a.comps[0][8 * N9 : -1] = 0
        chirp2a.instf[0][8 * N9 : -1] = 0
        chirp2a[4 * N9 : 5 * N9] = 0
        chirp2a.comps[0][4 * N9 : 5 * N9] = 0
        chirp2a.instf[0][4 * N9 : 5 * N9] = 0
        chirp2a[N9 : 4 * N9] = chirp2a[N9 : 4 * N9] * sg.windows.tukey(3 * N9, 0.25)
        chirp2a.comps[0][N9 : 4 * N9] = chirp2a[N9 : 4 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2a._instf[0][5 * N9 : 8 * N9] = 0
        chirp2a.comps[0][5 * N9 : 8 * N9] = 0
        chirp2a[5 * N9 : 8 * N9] = 0

        chirp2b[0:N9] = 0
        chirp2b.comps[0][0:N9] = 0
        chirp2b.instf[0][0:N9] = 0
        chirp2b[8 * N9 : -1] = 0
        chirp2b.comps[0][8 * N9 : -1] = 0
        chirp2b.instf[0][8 * N9 : -1] = 0
        chirp2b[4 * N9 : 5 * N9] = 0
        chirp2b.comps[0][4 * N9 : 5 * N9] = 0
        chirp2b.instf[0][4 * N9 : 5 * N9] = 0
        chirp2b[5 * N9 : 8 * N9] = chirp2b[5 * N9 : 8 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2b.comps[0][5 * N9 : 8 * N9] = chirp2b[5 * N9 : 8 * N9] * sg.windows.tukey(
            3 * N9, 0.25
        )
        chirp2b._instf[0][N9 : 4 * N9] = 0
        chirp2b.comps[0][N9 : 4 * N9] = 0
        chirp2b[N9 : 4 * N9] = 0

        signal = chirp1 + chirp2b + chirp2a + chirp3

        # TODO: The total number of comps should be generated automatically.
        if not self.return_signal:
            signal = signal.view(np.ndarray)
            # signal.total_ncomps = 4

        return signal

    # Other signals ----------------------------------------------------------------

    def signal_hermite_function(self, order=18, t0=0.5, f0=0.25):
        """Generates a round hermite function of a given order. The spectrogram of
        Hermite functions are given by an annular ridge in the time frequency plane, the
        center of which is given by (t0,f0).

        Args:
            order (int, optional): Order of the Hermite function. Defaults to 18.
            t0 (float, optional): Time coordinate of the center of the spectrogram.
            Defaults to 0.5.
            f0 (float, optional): Frequency coordinate of the center of the spectrogram.
            Defaults to 0.25.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        t0 = int(N * t0)
        t = np.arange(N) - t0
        signal = hermite_fun(N, order, t=t, T=np.sqrt(2 * N)) * np.cos(2 * pi * f0 * t)

        if self.return_signal:
            signal = signal.view(Signal)
            signal._instf = signal[:]

        return signal

    def signal_hermite_elipse(self, order=30, t0=0.5, f0=0.25):
        """Generates a non-round Hermite function of a given order. The spectrogram of
        Hermite functions are given by an elipsoidal ridge in the time frequency plane,
        the center of which is given by (t0,f0).

        Args:
            order (int, optional): Order of the Hermite function. Defaults to 18.
            t0 (float, optional): Time coordinate of the center of the spectrogram.
            Defaults to 0.5.
            f0 (float, optional): Frequency coordinate of the center of the spectrogram.
            Defaults to 0.25.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        t0 = int(N * t0)
        t = np.arange(N) - t0
        signal = hermite_fun(N, order, t=t, T=1.5 * np.sqrt(2 * N)) * np.cos(
            2 * pi * f0 * t
        )

        if self.return_signal:
            signal = signal.view(Signal)
            signal._instf = signal[:]

        return signal

    def signal_mc_triple_impulse(self, Nimpulses=3):
        """Generates three equispaced impulses in time.

        Args:
            Nimpulses (int, optional): Number of impulses. Defaults to 3.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        dloc = Nsub / (Nimpulses + 1)
        signal = np.zeros((N,))
        comps = []

        T = np.sqrt(N)
        t = np.arange(0, N)
        p = 1 / 6
        T2 = T * p
        k = (T2**2 + T**2) / (T * T2) ** 2
        imp_fun = lambda t0: np.exp(-pi / (T2**2) * (t - t0) ** 2) * np.cos(
            2 * pi * 0.25 * (t - t0)
        )

        for i in range(Nimpulses):
            impulse = np.zeros((N,))
            t0 = int((i + 1) * dloc) + tmin
            impulse = imp_fun(t0)
            comps.append(impulse.copy())
            signal += impulse

        if self.return_signal:
            signal = signal.view(Signal)
            signal._comps = comps
            signal._instf = comps

        return signal

    def signal_mc_impulses(self, Nimpulses=5):
        """Generates equispaced impulses in time.

        Args:
            Nimpulses (int, optional): Number of impulses. Defaults to 3.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        dloc = Nsub / (Nimpulses + 1)
        signal = np.zeros((N,))
        comps = []

        T = np.sqrt(N)
        t = np.arange(0, N)
        p = 1 / 6
        T2 = T * p
        k = (T2**2 + T**2) / (T * T2) ** 2
        imp_fun = lambda t0: np.exp(-pi / (T2**2) * (t - t0) ** 2) * np.cos(
            2 * pi * 0.25 * (t - t0)
        )

        for i in range(Nimpulses):
            impulse = np.zeros((N,))
            t0 = int((i + 1) * dloc) + tmin
            impulse = imp_fun(t0) * (i + 1)
            comps.append(impulse.copy())
            signal += impulse

        if self.return_signal:
            signal = signal.view(Signal)
            signal._comps = comps
            signal._instf = comps

        return signal

    def signal_mc_exp_chirps(self):
        """Generates a multicomponent signal comprising three exponential chirps.


        #     Returns:
        #         numpy.ndarray: Returns a numpy array with the signal.
        #"""
        N = self.N
        signal = np.zeros((N,))

        exponents = [4, 3, 2]
        finits = [self.fmin, 1.8 * self.fmin, 2.5 * self.fmin]
        fends = [0.3, 0.8, 1.2]  # [, 0.8]#, 1.2]
        ncomps = len(fends)

        max_freq = self.fmax

        for i in range(ncomps):
            aux = np.zeros((N,))
            _, instf, tmin, tmax = self._signal_exp_chirp(
                finit=finits[i], fend=fends[i], exponent=exponents[i], r_instf=True
            )

            instf2 = instf

            if instf[0] >= max_freq:
                break

            instf[np.where(instf2 > max_freq)] = 0
            instf2 = instf2[np.where(instf2 <= max_freq)]
            tukwin = sg.windows.tukey(len(instf2), 0.5)

            self.check_inst_freq(instf2)
            phase = np.cumsum(instf2)
            x = np.cos(2 * pi * phase)
            tukwin = sg.windows.tukey(len(x), 0.99)

            x = x * tukwin
            aux[tmin : tmin + len(x)] = x

            signal = signal + Signal(aux, instf=instf)

        signal = signal - np.mean(signal)
        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_damped_cos(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        t = np.arange(N) / N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)
        fmax = self.fmax

        signal = np.zeros((N,))
        instf0 = np.zeros_like(signal)
        instf1 = np.zeros_like(signal)
        instf2 = np.zeros_like(signal)

        omega = 5
        f0 = 0.32 + 0.00 * tsub / Nsub
        if0 = f0 + np.exp(np.log(0.25) * tsub / Nsub) * 0.1 * np.cos(
            2 * pi * omega * tsub / Nsub - pi * omega
        )

        if1 = self.fmin + 0.1 * tsub / Nsub
        # if2 = f2+0.02 + 0.02*np.cos(2*pi*omega*tsub/Nsub - pi*omega)
        # if0 = if0[np.where(if0<fmax)]
        # if1 = if1[np.where(if1<fmax)]
        # if2 = if2[np.where(if2<fmax)]

        phase0 = np.cumsum(if0)
        phase1 = np.cumsum(if1)
        # phase2 = np.cumsum(if2)

        x0 = np.zeros_like(signal)
        x1 = np.zeros_like(signal)
        # x2 = np.zeros_like(signal)

        eps = 1e-6
        t = np.arange(Nsub) + eps
        c = 1 / Nsub / 10
        prec = 10e-1  # Precision at sample N for the envelope.
        alfa = -np.log(prec * Nsub / ((Nsub - c) ** 2)) / Nsub
        e = np.exp(-alfa * t) * ((t - c) ** 2 / t)
        e[0] = 0
        e = e / np.max(np.abs(e))

        x0[tmin : tmin + len(phase0)] = np.cos(2 * pi * phase0) * sg.windows.tukey(
            len(phase0), 0.25
        )
        x1[tmin : tmin + len(phase1)] = np.cos(2 * pi * phase1) * sg.windows.tukey(
            len(phase1), 0.25
        )
        # x2[tmin:tmin+len(phase2)] = np.cos(2*pi*phase2)*sg.windows.tukey(len(phase2),0.25)

        instf0[tmin : tmin + len(if0)] = if0
        instf1[tmin : tmin + len(if1)] = if1
        # instf2[tmin:tmin+len(if2)] = if2

        chirp0 = Signal(x0, instf=instf0)
        chirp1 = Signal(x1, instf=instf1)
        # chirp2 = Signal(x2, instf=instf2)

        # signal = x0+x1+x2

        signal = chirp0 + chirp1

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal

    def signal_mc_damped_cos_2(self):
        """Generates a multicomponent signal with different types of components.

        Returns:
            numpy.ndarray: Returns a numpy array with the signal.
        """

        N = self.N
        tmin = self.tmin
        tmax = N - tmin
        Nsub = tmax - tmin
        tsub = np.arange(Nsub)

        signal = np.zeros((N,))
        instf0 = np.zeros_like(signal)
        instf1 = np.zeros_like(signal)

        omega = 5
        f0 = 0.3 + 0.00 * tsub / Nsub
        if0 = f0 + np.exp(np.log(0.25) * tsub / Nsub) * 0.14 * np.cos(
            2 * pi * omega * tsub / Nsub - pi * omega
        )

        if1 = self.fmin + 0.15 * tsub / Nsub

        phase0 = np.cumsum(if0)
        phase1 = np.cumsum(if1)

        x0 = np.zeros_like(signal)
        x1 = np.zeros_like(signal)

        a1 = 0.5 + 0.3 * np.cos(1.3 * pi * omega * tsub / Nsub - pi * omega)

        x0[tmin : tmin + len(phase0)] = np.cos(2 * pi * phase0) * sg.windows.tukey(
            len(phase0), 0.25
        )
        x1[tmin : tmin + len(phase1)] = (
            a1 * np.cos(2 * pi * phase1) * sg.windows.tukey(len(phase1), 0.25)
        )

        instf0[tmin : tmin + len(if0)] = if0
        instf1[tmin : tmin + len(if1)] = if1

        chirp0 = Signal(x0, instf=instf0)
        chirp1 = Signal(x1, instf=instf1)

        signal = chirp0 + chirp1

        if not self.return_signal:
            return signal.view(np.ndarray)
        return signal


def hermite_poly(t, n, return_all=False):
    """Generates a Hermite polynomial of order n on the vector t.

    Args:
        t (ndarray, float): A real valued vector on which compute the Hermite
        polynomials.
        n (int): Order of the Hermite polynomial.

    Returns:
        ndarray: Returns an array with the Hermite polynomial computed on t.
    """

    all_hp = np.zeros((n + 1, len(t)))

    if n == 0:
        hp = np.ones((len(t),))
        all_hp[0] = hp

    else:
        hp = 2 * t
        all_hp[0, :] = np.ones((len(t),))
        all_hp[1, :] = hp
        # if n >= 1:
        for i in range(2, n + 1):
            # hp = 2*t*hermite_poly(t,n-1) - 2*(n-1)*hermite_poly(t,n-2)
            hp = 2 * t * all_hp[i - 1] - 2 * (i - 1) * all_hp[i - 2]
            all_hp[i, :] = hp

    if not return_all:
        return hp
    else:
        return hp, all_hp


def hermite_fun(N, q, t=None, T=None, return_all=False):
    """Computes an Hermite function of order q, that consist in a centered Hermite
    polynomial multiplied by the squared-root of a centered Gaussian given by:
    exp(-pi(t/T)^2). The parameter T fixes the width of the Gaussian function.

    Args:
        N (int): Length of the function in samples
        q (int): Order of the Hermite polynomial.
        t (ndarray): Values on which compute the function. If None, uses a centered
        vector from -N//2 to N//2-1. Defaults to None.
        T (float): Scale of the Gaussian involved in the Hermite function. If None,
        N = sqrt(N). Defaults to None.

    Returns:
        _type_: _description_
    """

    if t is None:
        t = np.arange(N) - N // 2

    if T is None:
        T = np.sqrt(N)

    _, all_hp = hermite_poly(np.sqrt(2 * pi) * t / T, q - 1, return_all=True)
    gaussian_basic = np.exp(-pi * (t / T) ** 2) / np.sqrt(T)
    hfunc = np.zeros((q, len(t)))

    for i in range(q):
        Cnorm = np.sqrt(factorial(q - 1) * (2 ** (q - 1 - 0.5)))
        # gaussian_basic /= np.sum(gaussian_basic)
        hfunc[i] = gaussian_basic * all_hp[i] / Cnorm

    if not return_all:
        return hfunc[-1]
    else:
        return hfunc[-1], hfunc



"""mcms-benchs: A toolbox that provides a common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import seaborn as sns
from mcsm_benchs.Benchmark import Benchmark
import numpy as np
from matplotlib import pyplot as plt, ticker as mticker
import string
import os
import plotly.express as px
import scipy.stats as spst


class ResultsInterpreter:
    """This class takes a Benchmark-class object to produce a series of plots and tables
    summarizing the obtained results:

    Methods
    -------
    def get_benchmark_as_data_frame(self):

    """

    def __init__(self, a_benchmark):
        if not isinstance(a_benchmark, Benchmark):
            # self.results = a_benchmark.get_results_as_df()
            # else:
            raise ValueError("Input should be a Benchmark.\n")

        self.benchmark = a_benchmark
        self.task = a_benchmark.task
        self.methods_ids = a_benchmark.methods_ids
        self.N = a_benchmark.N
        self.repetitions = a_benchmark.repetitions
        self.snr_values = a_benchmark.SNRin
        self.signal_ids = a_benchmark.signal_ids
        self.methods_and_params_dic = a_benchmark.methods_and_params_dic

        # Number of elements to display in the plots.
        self.ncomparisons = sum(
            [len(self.methods_and_params_dic[m]) for m in self.methods_ids]
        )

        self.path_results = os.path.join("results")
        self.path_results_figures = os.path.join("results", "figures")

    # --------------------------------------------------------------------------------------

    def get_benchmark_as_data_frame(self, idx=0):
        """Returns a DataFrame with the raw data produced by the benchmark with the
        following format:
            -------------------------------------------------------------------------
            | Method | Parameter | Signal_id | Repetition | SNRin_1 | ... | SNRin_n |
            -------------------------------------------------------------------------

        Returns:
            DataFrame: Raw data of the comparisons.
        """
        df = self.benchmark.get_results_as_df()
        if type(df) == list:
            df = df[idx]

        return df

    # --------------------------------------------------------------------------------------

    def rearrange_data_frame(self, results=None, idx=0):
        """Rearrange DataFrame table for using Seaborn library.

        Args:
            results (DataFrame, optional): If not None, must receive the DataFrame
            produced by a Benchmark-class object using get_results_as_df(). If None,
            uses the Benchmark object given to the constructor of the Interpreter.
            Defaults to None.

        Returns:
            DataFrame: Rearranged DataFrame
        """
        if results is None:
            df = self.benchmark.get_results_as_df()
            if type(df) == list:
                df = df[idx]
        else:
            df = results

        aux_dic = dict()
        new_columns = df.columns.values[0:5].copy()

        new_columns[-1] = "QRF"

        for i in range(4, df.shape[1]):
            idx = [j for j in range(4)] + [i]
            df_aux = df.iloc[:, idx]
            df_aux.columns = new_columns
            aux_dic[df.columns[i]] = df_aux

        df3 = pd.concat(aux_dic, axis=0)
        df3 = df3.reset_index()
        df3 = df3.drop(labels="level_1", axis=1)
        df3.columns.values[0] = "SNRin"
        return df3

    # --------------------------------------------------------------------------------------

    def _get_df_means(self, df=None, idx=0):
        """Generates a DataFrame of mean results to .md file.

        Returns:
            str: String containing the table.
        """
        # if filename is None:
        # filename = 'results'
        if df is None:
            df = self.benchmark.get_results_as_df()
            if type(df) == list:
                df = df[idx]

        # Check if matrix values are bools:
        aux = df.iloc[:, 4::].to_numpy()
        if type(aux[0, 0]) is bool:
            df.iloc[:, 4::] = aux.astype(int)

        column_names = ["Method + Param"] + [col for col in df.columns.values[4::]]
        df_means = list()

        for signal_id in self.signal_ids:
            methods_names = list()
            snr_out_mean = np.zeros((1, len([col for col in df.columns.values[4::]])))
            snr_out_std = np.zeros((1, len([col for col in df.columns.values[4::]])))
            aux_dic = dict()

            df2 = df[df["Signal_id"] == signal_id]
            for metodo in self.methods_and_params_dic:
                tag = metodo
                aux = df2[df2["Method"] == metodo]
                if len(self.methods_and_params_dic[metodo]) > 1:
                    for params in self.methods_and_params_dic[metodo]:
                        methods_names.append(tag + params)
                        valores = df2[
                            (df2["Method"] == metodo) & (df2["Parameter"] == params)
                        ]
                        # Compute the means
                        valores_mean = valores.iloc[:, 4::].to_numpy().mean(axis=0)
                        valores_mean.resize((1, valores_mean.shape[0]))
                        snr_out_mean = np.concatenate((snr_out_mean, valores_mean))
                        # Compute the std
                        # valores_std = valores.iloc[:,4::].to_numpy().std(axis = 0)
                        # valores_std.resize((1,valores_std.shape[0]))
                        # snr_out_std = np.concatenate((snr_out_std,valores_std))

                else:
                    methods_names.append(tag)
                    valores = df2[df2["Method"] == metodo]
                    # Compute the means
                    valores_mean = valores.iloc[:, 4::].to_numpy().mean(axis=0)
                    valores_mean.resize((1, valores_mean.shape[0]))
                    snr_out_mean = np.concatenate((snr_out_mean, valores_mean))
                    # Compute the std
                    # valores_std = valores.iloc[:,4::].to_numpy().std(axis = 0)
                    # valores_std.resize((1,valores_std.shape[0]))
                    # snr_out_std = np.concatenate((snr_out_std,valores_std))

            snr_out_mean = snr_out_mean[1::]
            # snr_out_std = snr_out_std[1::]
            aux_dic[column_names[0]] = methods_names
            for i in range(1, len(column_names)):
                aux_dic[str(column_names[i])] = snr_out_mean[:, i - 1]

            df_means.append(pd.DataFrame(aux_dic))

        return df_means

    # ---------------------------------------------------------------------------------------

    def _get_df_std(self, df=None, varfun=None, idx=0):
        """Generates a DataFrame with a variability function from the results.
        Returns an string to create a .md file.

        Args:
            df (_type_, optional): Main DataFrame with the results. Defaults to None.
            varfun (str, optional): A variability function. Defaults to 'std', which uses numpy.std().
            idx (int, optional): Index of the corresponding performance function. Defaults to 0.

        Returns:
            Two DataFrames with the lower and upper variability indicator.
        """
        # Use the benchmark results if no DataFrame is passed as parameter.
        if df is None:
            df = self.benchmark.get_results_as_df()
            if type(df) == list:
                df = df[idx]

        # Choose an appropriate variability measure:
        if varfun is None:
            if self.task == "detection":  # Bonferroni corrected Clopper-Pearson CI
                varfun = lambda x: _clopper_pearson(
                    x,
                    alpha=0.05,
                    bonferroni=self.ncomparisons,
                )
            # For denoising or misc, use bootstrap CI.
            else:
                # varfun = lambda x: (np.std(x),np.std(x))
                varfun = lambda x: _ci(
                    x,
                    alpha=0.05,
                    bonferroni=self.ncomparisons,
                )

        # Check if matrix values are bools:
        aux = df.iloc[:, 4::].to_numpy()
        if type(aux[0, 0]) is bool:
            df.iloc[:, 4::] = aux.astype(int)

        column_names = ["Method + Param"] + [col for col in df.columns.values[4::]]
        output_string = ""
        df_std_upper = list()
        df_std_lower = list()

        for signal_id in self.signal_ids:
            methods_names = list()
            snr_out_std = np.zeros(
                (1, 2 * len([col for col in df.columns.values[4::]]))
            )
            aux_dic = dict()
            aux_dic_2 = dict()

            df2 = df[df["Signal_id"] == signal_id]
            for metodo in self.methods_and_params_dic:
                tag = metodo
                aux = df2[df2["Method"] == metodo]
                if len(self.methods_and_params_dic[metodo]) > 1:
                    for params in self.methods_and_params_dic[metodo]:
                        methods_names.append(tag + params)
                        valores = df2[
                            (df2["Method"] == metodo) & (df2["Parameter"] == params)
                        ]
                        # Compute the std
                        valores_std = np.array(
                            [varfun(col) for col in valores.iloc[:, 4::].to_numpy().T]
                        )
                        valores_std.resize((1, valores_std.size))
                        snr_out_std = np.concatenate((snr_out_std, valores_std))

                else:
                    methods_names.append(tag)
                    valores = df2[df2["Method"] == metodo]
                    # Compute the std
                    # valores_std = valores.iloc[:,4::].to_numpy().std(axis = 0)
                    valores_std = np.array(
                        [varfun(col) for col in valores.iloc[:, 4::].to_numpy().T]
                    )
                    valores_std.resize((1, valores_std.size))
                    snr_out_std = np.concatenate((snr_out_std, valores_std))

            snr_out_std = snr_out_std[1::]
            aux_dic[column_names[0]] = methods_names
            for i, j in zip(
                range(1, len(column_names)), range(0, snr_out_std.shape[1], 2)
            ):
                aux_dic[str(column_names[i])] = snr_out_std[:, j]

            df_std_lower.append(pd.DataFrame(aux_dic))

            aux_dic_2[column_names[0]] = methods_names
            for i, j in zip(
                range(1, len(column_names)), range(1, snr_out_std.shape[1], 2)
            ):
                aux_dic_2[str(column_names[i])] = snr_out_std[:, j]

            df_std_upper.append(pd.DataFrame(aux_dic_2))

        return df_std_upper, df_std_lower

    # --------------------------------------------------------------------------------------
    def _get_table_means_and_std(self, link="", pm_name=None, idx=0):
        """Generates a table of mean and std results to .md file.
        Highlights the best results.

        Returns:
            str: String containing the table.
        """
        task = self.benchmark.task
        df = self.benchmark.get_results_as_df()
        if type(df) == list:
            df = df[idx]

        column_names = ["Method + Param"] + [col for col in df.columns.values[4::]]
        output_string = ""

        # Get dataframes with means and some variability measure
        dfs_means = self._get_df_means()
        dfs_std_upper, dfs_std_lower = self._get_df_std()

        for dfu, dfl, dfm in zip(dfs_std_upper, dfs_std_lower, dfs_means):
            dfu.iloc[:, 1::] = dfu.iloc[:, 1::] + dfm.iloc[:, 1::]
            dfl.iloc[:, 1::] = dfm.iloc[:, 1::] - dfl.iloc[:, 1::]

        # TODO: Round std to 2 decimal places.
        # def round_to_2dec(input):
        # if type(input) is float:
        # return '{:.2f}'.format(input)

        # dfs_std_lower = [df.applymap(round_to_2dec) for df in dfs_std_lower]
        # dfs_std_upper = [df.applymap(round_to_2dec) for df in dfs_std_upper]

        dfs_std = [
            {
                col: [[low, upp] for low, upp in zip(dfl[col], dfu[col])]
                for col in dfu.columns.values[1::]
            }
            for dfu, dfl in zip(dfs_std_upper, dfs_std_lower)
        ]
        dfs_std = [pd.DataFrame(df) for df in dfs_std]

        for signal_id, df_means, df_std in zip(self.signal_ids, dfs_means, dfs_std):
            df_means_aux = df_means.copy()

            # Format all floats with two decimals:
            nparray_aux = df_means.iloc[:, 1::].to_numpy()
            maxinds = np.argmax(nparray_aux, axis=0)

            for col, max_ind in enumerate(maxinds):
                for i in range(
                    df_means_aux.shape[0]
                ):  # range(len(df_means.loc[col+1])):
                    df_means_aux.iloc[i, col + 1] = "{:.2f}".format(
                        df_means.iloc[i, col + 1]
                    )

                #  Also, highlight maxima
                df_means_aux.iloc[max_ind, col + 1] = (
                    "**" + "{:.2f}".format(df_means.iloc[max_ind, col + 1]) + "**"
                )

            # Change column names to make it more human-readable
            df_results = pd.DataFrame()
            df_results[column_names[0]] = df_means[column_names[0]]
            for col_ind in range(1, len(column_names)):
                df_results["SNRin=" + str(column_names[col_ind]) + "dB (average)"] = (
                    df_means_aux[str(column_names[col_ind])]
                )
                # df_results['SNRin='+str(column_names[col_ind])+'dB (CI)'] = df_std[str(column_names[col_ind])]

                # Make sure to have two decimals precision in each std column.
                df_results["SNRin=" + str(column_names[col_ind]) + "dB (CI)"] = [
                    ["{:.2f}".format(a) for a in row]
                    for row in df_std[str(column_names[col_ind])]
                ]

                # df_results['SNRin='+str(column_names[col_ind])+'dB (CI)'] = df_std[str(column_names[col_ind])]

            # Make sure having two decimals precision in each std column

            if pm_name is not None:
                df_results.rename(columns={"QRF": pm_name})

            # Table header with links
            aux_string = (
                "### Signal: "
                + signal_id
                + " [[View Plot]]("
                + link
                + "/plot_"
                + signal_id
                + ".html)  "
                + "  [[Get .csv]]("
                + link
                + "/results_"
                + signal_id
                + ".csv"
                + ")"
                + "\n"
                + df_results.to_markdown(floatfmt=".2f")
                + "\n"
            )
            output_string += aux_string

        return output_string

    # --------------------------------------------------------------------------------------
    def _get_report_preamble(self, link=None):
        """Creates the preamble of the .md file with a table summarizing the benchmark results.

        Returns:
            str: String with the table header.
        """

        if link is None:
            link_text = ""
        else:
            link_text = " [[Results .csv]](" + link + "/results.csv" + ") " + "\n"

        lines = [
            "# Benchmark Report" + link_text + "\n",
            "## Configuration" + "\n",
            # 'Parallelize' + str(self.benchmark.parallel_flag) + '\n',
            "Length of signals: " + str(self.N) + "\n",
            "Repetitions: " + str(self.repetitions) + "\n",
            "SNRin values: ",
        ]

        lines = lines + [str(val) + ", " for val in self.snr_values] + ["\n"]
        lines = (
            lines
            + ["### Methods  \n"]
            + ["* " + methid + " \n" for methid in self.methods_ids]
        )
        lines = (
            lines
            + ["### Signals  \n"]
            + ["* " + signid + " \n" for signid in self.signal_ids]
        )
        # lines = lines + ['## Figures:\n ![Summary of results](results/../figures/plots_grid.png) \n']
        lines = lines + ["## Mean results tables: \n"]

        # TODO Check this part for different perf. functions
        if self.task == "denoising":
            lines = lines + [
                "The results shown here are the average and 95\% CI of \
                              the performance metric with Bonferroni correction. \
                              Best performances are **bolded**. \n"
            ]

        if self.task == "detection":
            lines = lines + [
                "The results shown here are the average and 95\% Clopper-Pearson CI of \
                            the estimated detection power with Bonferroni correction. \
                            Best performances are **bolded**. \n"
            ]

        return lines

    # --------------------------------------------------------------------------------------
    def save_report(
        self, filename=None, path="results", bars=False, link="", pm_name=None
    ):
        """This function generates a report of the results given in the Benchmark-class
        object. The report is saved in a Markdown syntax to be viewed as a .md file,
        while a .csv file is generated with the results.

        Args:
            filename (str, optional): Path for saving the report. Defaults to None.
        """
        if filename is None:
            filename = "results_" + self.task + ".md"

        # Check path availability, otherwise create folder.
        if not os.path.exists(path):
            os.makedirs(path)

        output_path = os.path.join(path, filename)

        # Generate table header:
        lines = self._get_report_preamble(link=link)
        with open(output_path, "w") as f:
            f.write("\n".join(lines))
            # f.writelines(lines)

        # Append table under header
        table_string = self._get_table_means_and_std(link=link, pm_name=pm_name)
        with open(output_path, "a") as f:
            f.write(table_string)

        return True

    # -------------------------------------------------------------------------------------
    def get_snr_plot(
        self,
        df,
        x=None,
        y=None,
        hue=None,
        axis=None,
        ylabel=None,
        plot_type="linear",
        **kwargs
    ):
        """Generates a Quality Reconstruction Factor (QRF) vs. SNRin plot. The QRF is
        computed as:
        QRF = 20 log ( norm(x) / norm(x-x_r)) [dB]
        where x is the noiseless signal and x_r is de denoised estimation of x.

        Args:
            df (DataFrame): DataFrame with the results of the simulation.
            x (str, optional): Column name to use as the horizontal axis.
            Defaults to None.
            y (str, optional): Column name to use as the vertical axis.
            Defaults to None.
            hue (str, optional): Column name with the methods' name. Defaults to None.
            axis (matplotlib.Axes, optional): The axis object where the plot will be
            generated. Defaults to None.
        """

        markers = ["o", "d", "s", "*"]
        aux = np.unique(df[hue].to_numpy())
        # print(aux)
        # fig, axis2 = plt.subplots(1,1)

        plots = [
            (method_name, markers[np.mod(i, 4)]) for i, method_name in enumerate(aux)
        ]
        u_offset = np.linspace(-2, 2, len(plots))
        u_offset = np.zeros_like(u_offset)
        for offs_idx, plots_info in enumerate(plots):
            method_name, marker = plots_info
            df_aux = df[df[hue] == method_name]
            u = np.unique(df_aux[x].to_numpy())
            v = np.zeros_like(u, dtype=float)

            label = "".join(
                [c for c in string.capwords(method_name, sep="_") if c.isupper()]
            )
            if method_name.find("-") > -1:
                label = label + method_name[method_name.find("-") : :]

            for uind, j in enumerate(u):
                df_aux2 = df_aux[df_aux[x] == j]
                no_nans = df_aux2[y].dropna()
                if no_nans.shape != (0,):
                    v[uind] = np.nanmean(no_nans.to_numpy())

            if plot_type == "linear":
                axis.plot(
                    u + u_offset[offs_idx],
                    v,
                    "-" + marker,
                    # ms = 5,
                    # linewidth = 1.0,
                    label=label,
                    **kwargs
                )

            if plot_type == "semilogy":
                axis.semilogy(
                    u + u_offset[offs_idx],
                    v + 1e-10,
                    "-" + marker,
                    label=label,
                    **kwargs
                )
                axis.yaxis.set_minor_locator(
                    mticker.LogLocator(numticks=999, subs="auto")
                )

        # axis.plot([np.min(u), np.max(u)],[np.min(u), np.max(u)],'r',
        # linestyle = (0, (5, 10)),
        # linewidth = 0.75)
        axis.set_xticks(u)
        axis.set_yticks(v)
        # axis.set_xlabel(x + ' (dB)')
        # axis.set_ylabel(y + ' (dB)')

        axis.set_xlabel("SNRin (dB)")

        if ylabel is None:
            if self.benchmark.task == "denoising":
                axis.set_ylabel("QRF (dB)")

            if self.benchmark.task == "detection":
                axis.set_ylabel("Detection Power")
        else:
            axis.set_ylabel(ylabel)

        return True

    # --------------------------------------------------------------------------------------
    def get_snr_plot_bars(
        self,
        df,
        x=None,
        y=None,
        hue=None,
        errbar_fun=("ci", 95),
        axis=None,
        errbar_params=None,
    ):
        """Generates a Quality Reconstruction Factor (QRF) vs. SNRin barplot.
        The QRF is computed as:
                        QRF = 20 log ( norm(x) / norm(x-x_r)) [dB]
        where x is the noiseless signal and x_r is de denoised estimation of x.

        Args:
            df (DataFrame): DataFrame with the results of the simulation.
            x (str, optional): Column name to use as the horizontal axis.
            Defaults to None.
            y (str, optional): Column name to use as the vertical axis.
            Defaults to None.
            hue (str, optional): Column name with the methods' name. Defaults to None.
            axis (matplotlib.Axes, optional): The axis object where the plot will be
            generated. Defaults to None.
        """

        if errbar_params is None:
            errbar_params = {
                "errwidth": 0.1,
                "capsize": 0.02,
            }

        barfig = sns.barplot(
            x="SNRin",
            y="QRF",  # It's QRF in the DataFrame, but changed later.
            hue="Method",
            data=df,
            dodge=True,
            errorbar=errbar_fun,
            ax=axis,
            **errbar_params
        )

        axis.set_xlabel("SNRin (dB)")
        if self.benchmark.task == "denoising":
            axis.set_ylabel("QRF (dB)")

        if self.benchmark.task == "detection":
            axis.set_ylabel("Detection Power")

        return barfig

    # --------------------------------------------------------------------------------------
    def get_summary_plots(
        self,
        df_rearr=None,
        size=(3, 3),
        savetofile=True,
        path=None,
        filter_crit="all",
        filter_str=None,
        errbar_fun=("ci", 95),
        errbar_params=None,
        ax=None,
        plot_type="bars",
        bars=True,
        magnitude="absolute",
    ):
        """Generates individual performance plots for each signal, displaying the
        performance measure of all methods for all noise conditions.

        Args:
            size (tuple, optional): Size (in inches) of the figures. Defaults to (3,3).
            savetofile (bool, optional): Whether to save or not the figures.
            Defaults to True.
            filename (_type_, optional): Path and file name to save the figures. If None
            figures are saved in "results/figures" . Defaults to None.
            filter_str (_type_, optional): A string, or a list of strings, to select
            the methods to plot. If None, plots all the methods. Defaults to None.
            filter_crit (str, optional): A criterion to use the strings passed in
            filter_str. If 'all', only choose those methods where all the strings
            appear in the "Method" column of the resuls DataFrame. If 'any', select the methods for wich any of the strings appear in the mentioned column.
            Defaults to 'all'.

            plot_type (str, optional): _description_. Defaults to 'lines'.

        Returns:
            list: A list with matplotlib figures.
        """

        # Nsignals = len(self.signal_ids)

        if df_rearr is None:
            df_rearr = self.rearrange_data_frame()

        if magnitude == "difference":
            df_rearr["QRF"] = df_rearr["QRF"] - df_rearr["SNRin"]

        list_figs = list()

        for signal_id in self.signal_ids:
            if ax is None:
                fig, ax = plt.subplots(1, 1)
            else:
                fig = plt.gca()

            print(signal_id)
            # sns.set_theme()
            df_aux = df_rearr[df_rearr["Signal_id"] == signal_id]

            # If the method has different parameters, add the parameters to the name.
            indexes = df_aux["Parameter"] != "{(),{}}"
            method_and_params = (
                df_aux.loc[indexes, "Method"] + "-" + df_aux.loc[indexes, "Parameter"]
            )
            df_aux.loc[indexes, "Method"] = method_and_params

            # Filter methods using the string given:
            if filter_str is not None:
                if filter_crit == "all":
                    a = [np.all([j in i for j in filter_str]) for i in df_aux["Method"]]
                if filter_crit == "any":
                    a = [np.any([j in i for j in filter_str]) for i in df_aux["Method"]]

                method_and_params = df_aux.iloc[a, :]
                df_aux = method_and_params

            if bars == False or plot_type == "lines":
                self.get_snr_plot(df_aux, x="SNRin", y="QRF", hue="Method", axis=ax)
            else:
                self.get_snr_plot_bars(
                    df_aux,
                    x="SNRin",
                    y="QRF",
                    hue="Method",
                    errbar_fun=errbar_fun,
                    errbar_params=errbar_params,
                    axis=ax,
                )

            if self.benchmark.task == "denoising" and magnitude == "difference":
                ax.set_ylabel("QRF - SNR")

            if self.benchmark.task == "detection":
                ax.set_ylabel("Detection Power")
                ax.set_ylim([0, 1.2])

            ax.grid(linewidth=0.1)
            ax.set_axisbelow(True)
            ax.set_title(signal_id)
            ax.legend(loc="upper left", frameon=False, fontsize="small")
            fig.set_size_inches(size)

            # Save figures to file.
            if savetofile:
                if path is None:
                    filename = os.path.join(
                        "results", self.task, "figures", "plot_" + signal_id + ".pdf"
                    )

                else:
                    filename = os.path.join(path, "plot_" + signal_id + ".pdf")

                fig.savefig(filename, bbox_inches="tight")

            list_figs.append(fig)
            ax = None  # Create new figure in the next interation of the loop.
        return list_figs

    # --------------------------------------------------------------------------------------
    def get_summary_plotlys(
        self, df=None, bars=True, difference=False, varfun=None, ylabel=None, idx=0
    ):
        """Generates interactive plots with plotly.

        Returns:
            list : A list with plotly figures.
        """

        if df is None:
            df = self.benchmark.get_results_as_df()
            if type(df) == list:
                df = df[idx]

        if difference:
            for col in df.columns.values[4::]:
                df[col] = df[col] - col

        figs = []

        # Get dataframes with means and some variability measure (varfun)
        dfs_means = self._get_df_means(df=df)
        dfs_std_upper, dfs_std_lower = self._get_df_std(df=df, varfun=varfun)

        for signal_id, df_means, df_std_upper, df_std_lower in zip(
            self.signal_ids, dfs_means, dfs_std_upper, dfs_std_lower
        ):
            df3 = df_means.set_index("Method + Param").stack().reset_index()
            df3.rename(columns={"level_1": "SNRin", 0: "QRF"}, inplace=True)
            df3_std = df_std_upper.set_index("Method + Param").stack().reset_index()
            df3_std.rename(columns={"level_1": "SNRin", 0: "std"}, inplace=True)
            df3_std_minus = (
                df_std_lower.set_index("Method + Param").stack().reset_index()
            )
            df3_std_minus.rename(columns={"level_1": "SNRin", 0: "std"}, inplace=True)
            df3["std"] = df3_std["std"]
            df3["std-minus"] = df3_std_minus["std"]
            # print(df3)

            if bars:
                fig = px.bar(
                    df3,
                    x="SNRin",
                    y="QRF",
                    color="Method + Param",
                    #  markers=True,
                    barmode="group",
                    error_x="SNRin",
                    error_y="std",
                    error_y_minus="std-minus",
                    title=signal_id,
                )
            else:
                fig = px.line(
                    df3,
                    x="SNRin",
                    y="QRF",
                    color="Method + Param",
                    markers=True,
                    error_x="SNRin",
                    error_y="std",
                    error_y_minus="std-minus",
                    title=signal_id,
                )

            fig.update_layout(xaxis_title="SNRin (dB)")

            if ylabel is not None:
                fig.update_layout(yaxis_title=ylabel)

            figs.append(fig)

        return figs

    def get_html_figures(
        self,
        df=None,
        path=None,
        bars=True,
        difference=False,
        varfun=None,
        ylabel=None,
        idx=0,
    ):
        """
        Generate .html interactive plots files with plotly to show online.

        """
        if df is None:
            df = self.benchmark.get_results_as_df()
            if type(df) == list:
                df = df[idx]

        figs = self.get_summary_plotlys(
            df=df, bars=bars, difference=difference, varfun=varfun, ylabel=ylabel
        )

        for signal_id, fig in zip(self.signal_ids, figs):

            if path is None:
                path = "results"

            # Check path availability, otherwise create folder.
            if not os.path.exists(path):
                os.makedirs(path)

            filename = os.path.join(path, "plot_" + signal_id + ".html")
            with open(filename, "w") as f:
                f.write(fig.to_html(full_html=False, include_plotlyjs="cdn"))

        return True

    def get_csv_files(self, path, idx=0):
        """Generates a table of mean results to .md file.
        Saves a .csv file with the results per signal.

        Returns:
            str: String containing the table.
        """

        # TODO Deal with results for different perf. metrics here!
        df = self.benchmark.get_results_as_df()
        if type(df) == list:
            df = df[idx]

        # Save a .csv with all the results.
        df_aux = self.rearrange_data_frame()
        filename = os.path.join(path, "results.csv")
        df_aux.to_csv(filename)

        # Save one .csv per signal.
        for signal_id in self.signal_ids:
            # Generate DataFrame with only signal information
            df2 = df[df["Signal_id"] == signal_id]

            # # Save .csv file for the signal.
            if path is None:
                path = "results"

            # Check path availability, otherwise create folder.
            if not os.path.exists(path):
                os.makedirs(path)

            filename = os.path.join(path, "results_" + signal_id + ".csv")
            df2.to_csv(filename)

        return True

    def elapsed_time_summary(self):
        """Get a DataFrame summarizing the elapsed times of the methods.

        Returns:
            DataFrame: Pandas DataFrame with a summary of the elapsed times.
        """
        mydict = self.benchmark.elapsed_time
        auxdic = {}

        for signal_id in mydict:
            for k in mydict[signal_id]:
                for k2 in mydict[signal_id][k]:
                    auxdic[signal_id + "-" + k + "-" + k2] = pd.DataFrame(
                        mydict[signal_id][k][k2]
                    )

        df3 = pd.concat(auxdic, axis=0)
        df3 = df3.reset_index()
        df3 = df3.drop(labels="level_1", axis=1)
        methid = np.unique(df3["level_0"])
        auxdic = {}
        for i in methid:
            auxdic[i] = (
                np.mean(df3[0][df3["level_0"] == i]),
                np.std(df3[0][df3["level_0"] == i]),
            )

        df = pd.DataFrame(auxdic)
        df = df.transpose()
        df.columns = ("Average time (s)", "Std")

        return df


# Use this function only for the CP CI shown in the interactive figures using Plotly:
def _clopper_pearson(x, alpha=0.05, bonferroni=1):
    """
    Clopper-Pearson confidence interval for Bernoulli parameter
    alpha: confidence level
    k: number of successes
    n: number of observations
    """
    alpha = alpha / bonferroni
    n = len(x)  # k: number of successes
    k = sum(x)  # n: number of observations
    lb = spst.beta.ppf(alpha / 2, k, n - k + 1)
    ub = spst.beta.ppf(1 - alpha / 2, k + 1, n - k)

    if np.isnan(ub):
        ub = 1.0

    return np.mean(x) - lb, ub - np.mean(x)


def _ci(x, alpha=0.05, bonferroni=1):
    significance = alpha / bonferroni * 100
    lb, ub = sns.utils.ci(sns.algorithms.bootstrap(x), which=100 - significance)
    return np.mean(x) - lb, ub - np.mean(x)



"""mcms-benchs: A common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from abc import ABC, abstractmethod


class MethodTemplate(ABC):
    """An abstract class for new methods to add in a benchmark."""

    @abstractmethod
    def __init__(self): ...

    @abstractmethod
    def method(self): ...

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, task):
        self._task = task

    def get_parameters(self):
        return (((), {}),)



"""mcms-benchs: A toolbox that provides a common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
import numbers
import importlib


# Check oct2py is installed
try:
    octave_is_present = importlib.util.find_spec("oct2py")
    if octave_is_present:
        from oct2py import octave

except RuntimeError:
    print("Oct2py or Octave installation not found.")

import os


class OctaveInterface:
    """This class offers an interface between python and Matlab to seamlessly run methods in a Benchmark."""

    def __init__(self, octave_function_name, add2path=[], matlab_warnings=False):
        """Creates a new MatlabInterface method that calls a Matlab function.

        Args:
            matlab_function_name (str): The Matlab function name.
            add2path (list, optional): Add new paths where to look for the function indicated. Defaults to [].
            matlab_warnings (bool, optional): When True, prints out Matlab warnings. Defaults to False.

        Returns:
            MatlabInterface: An object able to call a function implemented in Matlab.
        """
        self.octave_function_name = octave_function_name
        self.eng = octave

        # try:
        #    self.eng = matlab.engine.start_matlab()
        # except NameError:
        #     print("Matlab engine or Matlab installation not found.")
        #     return None

        # if not matlab_warnings:
        #     self.eng.eval("warning('off','all');", nargout=0)

        octave.addpath(os.path.join("..", "src", "methods"))
        octave.addpath(os.path.join("src", "methods"))

        for path in add2path:
            octave.addpath(path)

    def octave_function(self, signal, *params):
        """A wrapper of a Matlab function that receives a signal to process and a variable number of positional arguments.

        Args:
            signal (numpy.ndarray): A numpy array with a signal.

        Returns:
            An equivalent array with the outputs of the Matlab function.
        """
        all_params = list((signal.copy(), *params))
        # params = self.pre_parameters(*all_params)
        fun_handler = getattr(self.eng, self.octave_function_name)
        outputs = fun_handler(*all_params)
        if isinstance(outputs, numbers.Number):
            return outputs

        if len(outputs) == 1:
            outputs = outputs[0]
        else:
            outputs = [output for output in outputs]
        return np.array(outputs)



"""mcms-benchs: A toolbox that provides a common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from mcsm_benchs.SignalBank import SignalBank, Signal
import pandas as pd
import numbers
import pickle

# from functools import partial
import multiprocessing

# from parallelbar import progress_map
import time

from numpy import mean, abs
from numpy.linalg import norm

from tqdm import tqdm


class Benchmark:
    """
    This class performs a number of tasks for methods comparison. It abstracts a benchmark itself.

    """

    def __init__(
        self,
        task="denoising",
        methods=None,
        N=256,
        Nsub=None,
        parameters=None,
        SNRin=None,
        repetitions=None,
        signal_ids="all",
        verbosity=1,
        parallelize=False,
        complex_noise=False,
        obj_fun=None,
        write_log=True,
        name=None,
        description=None,
        **kwargs,
    ):
        """Initialize the main parameters of the test bench before running the benchmark.

        Args:
            task (str, optional): The task to test the methods. Defaults to 'denoising'.

            methods (dict, optional): A dictionary of functions. Defaults to None.

            N (int, optional): Lengths of the observation window for the signal
            generation. Defaults to 256.

            Nsub (int, optional): Lengths of the signal. If None, uses Nsub (See
            SignalBank class description for details). Defaults to None.

            parameters (dict, optional): A dictionary of parameters for the methods
            to run. The keys of this dictionary must be same as the methods dictionary.
            Defaults to None.

            SNRin (tuple, optional): List or tuple with the SNR values.
            Defaults to None.

            repetitions (int, optional): Number of times each method is applied for
            each value of SNR. This value is the number of noise realizations that are
            used to assess the methods. Defaults to None.

            signal_ids (tuple, optional): Tuple or list of the signal ids from the
            SignalBank class. Defaults to 'all'.

            verbosity (int, optional): Number from 0 to 4. It determines the number of
            messages passed to the console informing the progress of the benchmarking
            process. Defaults to 1.

            parallelize (bool, optional): If True, tries to run the process in parallel.
            Defaults to False.

            complex_noise (bool, optional): If True, uses complex noise. A function can
            be passed to compute a new realization of noise instead.
            Defaults to False.

            obj_fun (callable, optional): If None, used the default objective functions
            for benchmarking. If a function is passed as an argument, the default is
            overridden.

            write_log (bool, optional): If True, saves a log of errors and warnings. Defaults to True.

            name (str, optional): A name to identify the benchmark in collaborative repositories.

            description (str, optional): A description to show in collaborative repositories.

        """

        self.write_log = write_log
        self.log = []

        self.name = name
        self.description = description

        # Check input parameters and initialize the object attributes
        self.input_parsing(
            task,
            methods,
            N,
            Nsub,
            parameters,
            SNRin,
            repetitions,
            signal_ids,
            verbosity,
            parallelize,
            complex_noise,
            obj_fun,
            **kwargs,
        )

        # Parallelize parameters
        if self.parallel_flag:
            if self.verbosity > 1:
                print("Number of processors: ", multiprocessing.cpu_count())
                print("Parallel pool: {}".format(self.processes))

    def __add__(self, other):
        """Overload the addition operator for Benchmark objects with the same parameters.

        Args:
            other (Benchmark): A Benchmark with the same parameters and possibly different methods.

        Raises:
            TypeError:

        Returns:
            Benchmark: A Benchmark object with the combined methods.
        """
        if isinstance(other, Benchmark):
            return self.sum(self, other)
        else:
            raise TypeError(
                "Unsupported operand type(s) for +: '{}' and '{}'".format(
                    type(self), type(other)
                )
            )

    def input_parsing(
        self,
        task,
        methods,
        N,
        Nsub,
        parameters,
        SNRin,
        repetitions,
        signal_ids,
        verbosity,
        parallelize,
        complex_noise,
        obj_fun,
        **kwargs,
    ):
        """Parse input parameters of the constructor of class Benchmark.

        Args:
            task (str, optional): The task to test the methods. Defaults to 'denoising'.
            methods (dict, optional): A dictionary of functions. Defaults to None.
            N (int, optional): Lengths of the signals. Defaults to 256.
            parameters (dict, optional): A dictionary of parameters for the methods
            to run. The keys of this dictionary must be same as the methods dictionary.
            Defaults to None.
            SNRin (tuple, optional): List or tuple with the SNR values.
            Defaults to None.
            repetitions (int, optional): Number of times each method is applied for
            each value of SNR.
            This value is the number of noise realizations that are used to assess the
            methods.Defaults to None.
            signal_ids (tuple, optional): Tuple or list of the signal ids from the
            SignalBank class. Defaults to 'all'.
            verbosity (int, optional): Number from 0 to 4. It determines the number of
            messages passed to the console informing the progress of the benchmarking
            process. Defaults to 1.
            parallelize (bool, optional): If True, tries to run the process in parallel.
            Defaults to False.

        Raises:
            ValueError: If any parameter is not correctly parsed.
        """
        # Check verbosity
        assert (
            isinstance(verbosity, int) and 0 <= verbosity < 6
        ), "Verbosity should be an integer between 0 and 5"
        self.verbosity = verbosity

        # TODO: ADD NEW TASKS
        # Check the task is either 'denoising' or 'detection'.
        # if (task != 'denoising' and task != 'detection'):
        #     raise ValueError("The tasks should be either 'denoising' or 'detection'.\n")
        # else:
        self.task = task

        # Check methods is a dictionary
        if type(methods) is not dict:
            raise ValueError("Methods should be a dictionary.\n")
        else:
            self.methods = methods
            self.methods_ids = [llave for llave in methods]

        # If no parameters are given to the benchmark.
        if parameters is None:
            self.parameters = {key: (((), {}),) for key in methods.keys()}
        else:
            if type(parameters) is dict:
                self.parameters = parameters
            else:
                raise ValueError("Parameters should be a dictionary or None.\n")

        # Check both dictionaries have the same keys:
        if not (self.methods.keys() == self.parameters.keys()):
            # sys.stderr.write
            raise ValueError(
                "Both methods and parameters dictionaries should have the same keys.\n"
            )

        # Check if N is an entire:
        if type(N) is int:
            self.N = N
        else:
            raise ValueError("N should be an entire.\n")

        # Check if Nsub is an entire:
        if Nsub is not None:
            if type(Nsub) is not int:
                raise ValueError("Nsub should be an entire.\n")

            # Check if Nsub is lower than N:
            if self.N < Nsub:
                raise ValueError("Nsub should be lower than N.\n")

        self.Nsub = Nsub

        # Check if SNRin is a tuple or list, and if so, check if there are only numerical variables.
        if (type(SNRin) is tuple) or (type(SNRin) is list):
            for i in SNRin:
                if not isinstance(i, numbers.Number):
                    raise ValueError("All elements in SNRin should be real numbers.\n")

            self.SNRin = SNRin
        else:
            raise ValueError("SNRin should be a tuple or a list.\n")

        # Check if repetitions is an entire:
        if type(repetitions) is int:
            self.repetitions = repetitions
        else:
            raise ValueError("Repetitions should be an entire.\n")

        # Check what to do with the signals:

        # Check if signal_ids is a dict with testing signals
        if isinstance(signal_ids, dict):
            signal_dic = {}
            for key in signal_ids:
                s = signal_ids[key]
                assert N == len(s), "Input signal length should be N"
                signal_dic[key] = Signal(s)

            self.signal_dic = signal_dic
            self.signal_ids = signal_ids
            self.tmin = 0  # int(np.sqrt(N))
            self.tmax = N  # -int(np.sqrt(N))

        else:
            if signal_ids == "all":
                # Generates a dictionary of signals
                signal_bank = SignalBank(N=self.N, Nsub=self.Nsub, return_signal=True)
                self.tmin = signal_bank.tmin  # Save initial and end times of signals.
                self.tmax = signal_bank.tmax

                self.signal_dic = signal_bank.signalDict
                self.signal_ids = [akey for akey in self.signal_dic]
                if Nsub is None:
                    self.Nsub = signal_bank.Nsub

            else:
                # Check if list of signals are in SignalBank
                if isinstance(signal_ids, tuple) or isinstance(signal_ids, list):
                    signal_bank = SignalBank(
                        N=self.N, Nsub=self.Nsub, return_signal=True
                    )
                    llaves = signal_bank.signalDict.keys()
                    assert all(signal_id in llaves for signal_id in signal_ids)

                self.signal_dic = signal_bank.signalDict
                self.tmin = signal_bank.tmin  # Save initial and end times of signals.
                self.tmax = signal_bank.tmax
                self.signal_ids = signal_ids

        # Check if complex_noise flag is bool:
        if type(complex_noise) is bool:
            self.complex_noise = complex_noise
        else:
            if callable(complex_noise):
                self.complex_noise = complex_noise
            else:
                if complex_noise != "NF":
                    raise ValueError("'complex_noise' should be a bool or callable.\n")

        # Handle parallelization parameters:
        max_processes = multiprocessing.cpu_count()

        if parallelize is False:
            self.parallel_flag = False
        else:
            if parallelize is True:
                if parallelize:
                    available_proc = multiprocessing.cpu_count()
                    self.processes = np.max((1, available_proc // 2))
                    self.parallel_flag = True
            else:
                if isinstance(parallelize, int):
                    if parallelize < max_processes:
                        self.processes = parallelize
                    else:
                        self.processes = max_processes
                    self.parallel_flag = True

        # Check objective function
        # Set the default performance function according to the selected task
        if obj_fun is None:
            self.objectiveFunction = {
                "perf_metric": self.set_objective_function(task),
            }
        else:
            if callable(obj_fun):
                self.objectiveFunction = {
                    "obj_fun": obj_fun,
                }
            else:
                if type(obj_fun) is dict:
                    self.objectiveFunction = obj_fun
                else:
                    raise ValueError(
                        "'obj_fun' should be a callable object or a dictionary.\n"
                    )

        # TODO Check this list of attribute initialization
        # Extra arguments may be passed when opening a saved benchmark:
        if kwargs == {}:
            # If we are here, this is a new benchmark, so the all the methods are new:
            self.this_method_is_new = {method: True for method in self.methods_ids}
            self.results = None
            self.noise_matrix = None
            self.methods_and_params_dic = dict()

        else:
            for key in kwargs:
                if not key in self.__dict__:
                    self.__dict__[key] = kwargs[key]

    def check_methods_output(self, output, input):
        """Check that the outputs of the method to benchmark fulfill the required type
        and shape.

        Args:
            output: Output from the method. The type and shape depends on the task.
            input: Input passed to the method to produce output.

        Raises:
            ValueError: If the output does not comply with the required type and shape
            for the selected task.
        """
        if self.task == "denoising":
            if type(output) is not np.ndarray:
                raise ValueError(
                    "Method's output should be a numpy array for task='denoising'.\n"
                )

            if output.shape != input.shape:
                raise ValueError(
                    "Method's output should have the same shape as input for task='denoising'.\n"
                )

    def set_objective_function(self, task):
        """
        Set the performance function for the selected task (future tasks could easily add new performance functions)
        """

        compFuncs = {  #'denoising': lambda x: self.snr_comparison(*x,tmin=self.tmin,tmax=self.tmax),
            "denoising": self.snr_comparison,
            "detection": self.detection_perf_function,
            "component_denoising": self.compare_qrf_block,
            "inst_frequency": self.compare_instf_block,
            "misc": None,
        }
        return compFuncs[task]

    def inner_loop(self, benchmark_parameters, timer=False):
        """Main loop of the Benchmark.

        Args:
            benchmark_parameters (tuple): Tuple or list with the parameters of the benchmark.
            timer (bool): If true, measures the time of execution.

        Returns:
            narray: Return a numpy array, the shape of which depends on the selected task.
        """

        method, params, idx = benchmark_parameters

        if self.verbosity >= 5:
            print("------ Inner loop. " + method + ": " + str(idx), flush=True)

        # Get the noisy signal (as a ndarray) and wrap it with the Signal class, adding
        # the signal information from the base signal.
        # This wrapper class behaves like a numpy array, but encapsulates signal info,
        # like the total number of components or number of components per time.

        noisy_signal = Signal(self.noisy_signals[idx])
        noisy_signal.ncomps = self.base_signal_info["ncomps"]
        noisy_signal.total_comps = self.base_signal_info["total_comps"]

        try:
            args, kwargs = params
            tstart = time.time()
            method_output = self.methods[method](noisy_signal, *args, **kwargs)
            elapsed = time.time() - tstart

        except BaseException as err:
            print(
                f"Unexpected error {err=}, {type(err)=} in method {method}. Watch out for NaN values."
            )

            # Write an entry log for errors.
            if self.write_log:
                log_entry = f"Unexpected error {err=}, {type(err)=} in method {method}. Noise matrix index: {idx=}."
                self.log.append(log_entry)

            elapsed = np.nan

            if self.task == "denoising":
                method_output = np.empty(noisy_signal.shape)
                method_output[:] = np.nan

            if self.task == "detection":
                method_output = np.nan

            if self.task == "misc":
                method_output = np.nan

        #! Rewrite this part.
        # self.check_methods_output(method_output,noisy_signals) # Just checking if the output its valid.

        return method_output, elapsed

    def set_results_dict(self):
        """Initializes a dictionary where the results are going to be saved later."""
        # If its a new benchmark...
        if self.results is None:
            self.results = dict()
            self.elapsed_time = dict()

            # Create nested dictionary for results and elapsed time:
            for fun_name in self.objectiveFunction:
                self.results[fun_name] = {}
                for signal_id in self.signal_ids:
                    self.results[fun_name][signal_id] = {}
                    self.elapsed_time[signal_id] = {}
                    for SNR in self.SNRin:
                        self.results[fun_name][signal_id][SNR] = {}
                        for method in self.methods:
                            self.results[fun_name][signal_id][SNR][method] = {}
                            self.elapsed_time[signal_id][method] = {}
                            # for param in self.parameters[method]:
                            #     self.results[signal_id][SNR][method][str(param)] = {}
        else:
            # Add new methods to the dictionary of results
            print("- Rerun benchmark.")
            print('-New methods',[method for method in self.methods if method not in self.results[self.objectiveFunction.keys()[0]][self.signal_ids.keys()[0]][self.SNRin[0]].keys()])

            for fun_name in self.objectiveFunction:
                for signal_id in self.signal_ids:
                    for SNR in self.SNRin:
                        for method in self.methods:
                            if (
                                method
                                not in self.results[fun_name][signal_id][SNR].keys()
                            ):
                                # print('-- New Method',method)
                                self.results[fun_name][signal_id][SNR][method] = {}
                                self.elapsed_time[signal_id][method] = {}

    def run_test(self): ## Make this a deprecation warning.
        print(
            "Method run_test() will be deprecated in newer versions. Use run() instead."
        )
        self.run()

    def run(self):
        """Run the benchmark.

        Returns:
            dict: Returns nested dictionaries with the results of the benchmark.
        """
        if self.verbosity >= 0:
            print("Running benchmark...")
            if self.verbosity > 1:
                bar_fun = lambda smth: smth
            else:
                bar_fun = lambda smth: tqdm(smth)

        # If it is a new benchmark, create a nested dict() to save the results.
        # If it is an old benchmark with new methods, create the new keys.
        self.set_results_dict()

        # This run all the experiments and save the results in nested dictionaries.
        for signal_id in self.signal_ids:
            if self.verbosity >= 1:
                print("- Signal " + signal_id)

            self.base_signal = self.signal_dic[signal_id]
            self.base_signal_info = self.signal_dic[signal_id].get_info()

            for SNR in bar_fun(self.SNRin):
                if self.verbosity >= 2:
                    print("-- SNR: {} dB".format(SNR))

                # If the benchmark has been run before,
                # re-run again with the same noise.
                if self.noise_matrix is None:
                    self.noise_matrix = self.generate_noise()

                noisy_signals, scaled_noise = self.sigmerge(
                    self.base_signal,
                    self.noise_matrix,
                    SNR,
                    tmin=self.tmin,
                    tmax=self.tmax,
                    return_noise=True,
                )

                # Access current noisy signals from the main loop.
                self.noisy_signals = noisy_signals

                # ===========================MAIN LOOP===================================

                # ------------------------- Parallel loop ------------------------------
                if self.parallel_flag:
                    parallel_list = list()
                    for method in self.methods:
                        if self.this_method_is_new[method]:
                            if self.verbosity >= 2:
                                print(
                                    "--- Parallel loop -- Method: "
                                    + method
                                    + "(all parameters)"
                                )
                            for p, params in enumerate(self.parameters[method]):
                                args, kwargs = get_args_and_kwargs(params)
                                for idx, noisy_signal in enumerate(noisy_signals):
                                    parallel_list.append([method, (args, kwargs), idx])

                    # Here implement the parallel stuff
                    pool = multiprocessing.Pool(processes=self.processes)
                    parallel_results = pool.map(self.inner_loop, parallel_list)
                    pool.close()
                    pool.join()
                    if self.verbosity >= 2:
                        print("--- Parallel loop finished.")

                # ---------------------- Serial loop -----------------------------------
                k = 0  # This is used to get the parallel results if it's necessary.

                for method in self.methods:

                    if self.this_method_is_new[method]:
                        if self.verbosity >= 3:
                            print("--- Method: " + method)

                        for p, params in enumerate(self.parameters[method]):
                            elapsed = []

                            if self.verbosity >= 4:
                                print("---- Parameters Combination: " + str(p))

                            args, kwargs = get_args_and_kwargs(params)

                            if self.task == "component_denoising":
                                extrargs = {"tmin": self.tmin, "tmax": self.tmax}
                                method_output = [
                                    [] for aaa in range(noisy_signals.shape[0])
                                ]

                            if self.task == "inst_frequency":
                                extrargs = {"tmin": self.tmin, "tmax": self.tmax}
                                # method_output = [[] for aaa in range(noisy_signals.shape[0])]
                                method_output = []

                            if self.task == "denoising":
                                extrargs = {"tmin": self.tmin, "tmax": self.tmax}
                                # method_output = np.zeros_like(noisy_signals)
                                method_output = []
                            if self.task == "detection":
                                extrargs = {}
                                # method_output = np.zeros((self.repetitions)).astype(bool)
                                method_output = []

                            if self.task == "misc":
                                method_output = []

                            for idx, noisy_signal in enumerate(noisy_signals):
                                # Get results from parallel computation
                                if self.parallel_flag:
                                    tmp, extime = parallel_results[k]
                                    # method_output[idx] = tmp
                                    method_output.append(tmp)
                                    # Save but DON'T TRUST the exec. time in parallel.
                                    elapsed.append(extime)
                                    k += 1

                                # Or from serial computation
                                else:
                                    tmp, extime = self.inner_loop(
                                        [method, (args, kwargs), idx]
                                    )
                                    # method_output[idx] = tmp
                                    method_output.append(tmp)
                                    elapsed.append(extime)

                            # Either way, results are saved in a nested dictionary-----
                            # For each performance metric
                            for fun_name in self.objectiveFunction:
                                result = []
                                for i, output in enumerate(method_output):
                                    try:
                                        fun = self.objectiveFunction[fun_name]
                                        perf_met_output = fun(
                                            self.base_signal,
                                            output,
                                            tmin=self.tmin,
                                            tmax=self.tmax,
                                            scaled_noise=scaled_noise[i],
                                        )

                                    except BaseException as err:
                                        print(
                                            f"Unexpected error {err=}, {type(err)=} while applying performance metric:{fun_name}. Watch out for NaN values."
                                        )
                                        perf_met_output = np.nan

                                    result.append(perf_met_output)

                                # Saving results -----------
                                if np.any([type(r) == dict for r in result]):
                                    aux = {key: [] for key in result[0]}
                                    for r in result:
                                        for key in r:
                                            aux[key].append(r[key])
                                    result = aux

                                self.results[fun_name][signal_id][SNR][method][
                                    str(params)
                                ] = result
                                self.elapsed_time[signal_id][method][
                                    str(params)
                                ] = elapsed

                            if self.verbosity > 4:
                                print("Elapsed:{}".format(np.mean(elapsed)))

                        self.methods_and_params_dic[method] = [
                            str(key) for key in self.parameters[method]
                        ]

        if self.verbosity > 0:
            print("The test has finished.")

        # Don't use old methods if run again.
        for method in self.this_method_is_new:
            self.this_method_is_new[method] = False

        return self.results

    def save_to_file(self, filename=None):
        """Save the results to a binary file that encodes the benchmark object.
        Notice that the methods associated with the benchmark, not being pickable objects,
        are NOT saved.

        Args:
            filename (str, optional): Path and filename. Defaults to None.

        Returns:
            bool: True if the file was successfully created.
        """

        if filename is None:
            filename = "a_benchmark"

        a_copy = self
        a_copy.methods = {key: None for key in a_copy.methods}
        a_copy.base_signal = a_copy.base_signal.view(np.ndarray)
        # a_copy.signal_dic = []
        a_copy.noisy_signals = a_copy.noisy_signals.view(np.ndarray)
        a_copy.objectiveFunction = {key: None for key in a_copy.objectiveFunction}

        if callable(a_copy.complex_noise):
            a_copy.complex_noise = "NF"

        with open(filename + ".pkl", "wb") as f:
            pickle.dump(a_copy.__dict__, f, protocol=pickle.HIGHEST_PROTOCOL)

        return True

    def get_results_as_df(self, results=None):
        """Get a pandas DataFrame object with the results of the benchmark.

        Args:
            results (dict, optional): Nested dictionary with the results of the
            benchmark. Defaults to None.

        Returns:
            DataFrame: Returns a pandas DataFrame with the results.
        """
        if list(self.results.keys())[0] in self.signal_ids:
            self.results = {"perf_metric": self.results}

        if results is None:
            df = self.dic2df(self.results)
        else:
            df = self.dic2df(results)

        df = pd.concat({param: df[param] for param in df.columns})
        df = df.unstack(level=3)
        df = df.reset_index()
        df.columns = pd.Index(
            ["Parameter", "Perf.Metric", "Signal_id", "Method", "Repetition"]
            + self.SNRin
        )
        df = df.reindex(
            columns=["Perf.Metric", "Method", "Parameter", "Signal_id", "Repetition"]
            + self.SNRin
        )
        df = df.sort_values(by=["Method", "Parameter", "Signal_id"])

        data_frames = []
        for per_fun in np.unique(df["Perf.Metric"]):
            dfaux = df[df["Perf.Metric"] == per_fun].iloc[:, 1::]
            aux2 = np.zeros((dfaux.shape[0],), dtype=bool)
            for metodo in self.methods_and_params_dic:
                aux = np.zeros((dfaux.shape[0],), dtype=bool)
                for params in self.methods_and_params_dic[metodo]:
                    aux = aux | (
                        (dfaux["Parameter"] == params) & (dfaux["Method"] == metodo)
                    )
                aux2 = aux2 | aux

            df2 = dfaux[aux2]
            data_frames.append(df2)

        if len(data_frames) == 1:
            return data_frames[0]
        else:
            return data_frames

    def add_new_method(self, methods, parameters=None, perf_func=None):
        """Add new methods to an existing Benchmark.

        Args:
            methods (_type_): A dictionary of methods.
            parameters (_type_, optional): If necessary, a dictionary of parameters for the new methods, or new parameters to explore with already benchmarked methods. Defaults to None.
        """
        # Check methods is a dictionary and update existing dictionary of methods.
        if type(methods) is not dict:
            raise ValueError("Methods should be a dictionary.\n")

        # If no parameters are given to the benchmark.
        if parameters is None:
            parameters = {key: (((), {}),) for key in methods}
        else:
            if type(parameters) is not dict:
                raise ValueError("Parameters should be a dictionary or None.\n")

        for key in methods:
            if key not in self.methods.keys():
                self.methods[key] = methods[key]
                self.methods_ids.append(key)
                self.parameters[key] = parameters[key]
                # self.elapsed_time[key]  = dict()
                self.this_method_is_new[key] = True

        # Check both dictionaries have the same keys:
        if not (self.methods.keys() == self.parameters.keys()):
            # sys.stderr.write
            raise ValueError(
                "Both methods and parameters dictionaries should have the same keys.\n"
            )

        self.objectiveFunction = perf_func
        # New methods cannot be parallelized (for now).
        self.parallel_flag = False

    # Other functions:
    def dic2df(self, mydic):
        """Get a pandas DataFrame from a nested dictionary.

        Args:
            mydic (dict): A nested dictionary of results coming from a Benchmark object.

        Returns:
            DataFrame : A DataFrame with a column per-level in the dictionary.
        """
        auxdic = dict()
        for key in mydic:
            if isinstance(mydic[key], dict):
                df = self.dic2df(mydic[key])
                auxdic[key] = df
            else:
                return pd.DataFrame(mydic)

        df = pd.concat(auxdic, axis=0)
        return df

    def generate_noise(self):
        """Generate a matrix of noise realizations of shape [self.repetitions,self.N].
        One realization per row.

        Returns:
            numpy.array: Matrix with noise realizations.
        """

        if isinstance(self.complex_noise, bool):
            np.random.seed(0)
            noise_matrix = np.random.randn(self.repetitions, self.N)
            if self.complex_noise:
                noise_matrix = np.random.rand(
                    self.repetitions, self.N
                ) + 1j * np.random.randn(self.repetitions, self.N)

        if callable(self.complex_noise):
            noise_matrix = np.random.randn(self.repetitions, self.N)
            for i in range(self.repetitions):
                noise_matrix[i, :] = self.complex_noise(self.N)

        return noise_matrix

    # Static methods--------------------------------------------------------------------
    @staticmethod
    def load_benchmark(filename, **kwargs):
        """Load a Benchmark object from a file saved using the class method.

        Args:
            filename (str): A path to the saved benchmark.

        Returns:
            Benchmark: A Benchmark object.
        """
        with open(filename + ".pkl", "rb") as f:
            benchmark_dict = pickle.load(f)

        return Benchmark(**benchmark_dict)

    @staticmethod
    def sigmerge(x1, noise, snr, tmin=None, tmax=None, return_noise=False):
        """Merge a signal and a noise realization with a given SNR (in dB).

        Args:
            x1 (_type_): _description_
            noise (_type_): _description_
            ratio (_type_): _description_
            tmin (_type_, optional): _description_. Defaults to None.
            tmax (_type_, optional): _description_. Defaults to None.
            return_noise (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_

        Example:
            >>> import mcsm_benchs.Benchmark
            >>> import mcsm_benchs.SignalBank
            >>> sb = SignalBank(N=1024)
            >>> signal = sb.signal_linear_chirp()
            >>> noise = np.random.randn(1024,)
            >>> snr = 10 # In dB
            >>> mixture = Benchmark.sigmerge(signal,noise,snr=snr)
        """
        # Get signal parameters.
        N = len(x1)

        if tmin is None:
            tmin = 0
        if tmax is None:
            tmax = N

        sig = np.random.randn(*noise.shape)
        ex1 = np.mean(np.abs(x1[tmin:tmax]) ** 2)

        if len(noise.shape) == 1:
            ex2 = np.mean(np.abs(noise) ** 2)
        else:
            ex2 = np.mean(np.abs(noise) ** 2, axis=1)
        h = np.sqrt(ex1 / (ex2 * 10 ** (snr / 10)))

        if len(noise.shape) > 1:
            h.resize((noise.shape[0], 1))

        scaled_noise = noise * h
        # sig = sig*h.reshape((noise.shape[0],1))
        sig = x1 + scaled_noise

        if return_noise:
            return sig, scaled_noise
        else:
            return sig

    @staticmethod
    def snr_comparison(x, x_hat, tmin=None, tmax=None, **kwargs):
        """
        Quality reconstruction factor for denoising performance characterization.
        """
        if tmin is None:
            tmin = 0

        if tmax is None:
            tmax = len(x)

        x = x[tmin:tmax]

        if len(x_hat.shape) == 1:
            x_hat = x_hat[tmin:tmax]
            qrf = 10 * np.log10(np.sum(x**2) / np.sum((x_hat - x) ** 2))
        else:
            x_hat = x_hat[:, tmin:tmax]
            qrf = np.zeros((x_hat.shape[0],))
            for i in range(x_hat.shape[0]):
                qrf[i] = 10 * np.log10(np.sum(x**2) / np.sum((x_hat[i, :] - x) ** 2))

        return qrf

    @staticmethod
    def detection_perf_function(original_signal, detection_output, **kwargs):
        """Performance function for detection. Returns the boolean output of the detection methods."""
        return detection_output

    @staticmethod
    def compare_qrf_block(signal, method_output, tmin=None, tmax=None, **kwargs):
        """ Compare the output SNR, i.e. qualitiy reconstruction factor (QRF) in a vectorized way for multi-component outputs.

        Args:
            signal (numpy.array or Signal): _description_
            method_output (numpy.array): _description_
            tmin (int, optional): tmin and tmax define a window of the signal to compare SNR. Defaults to None.
            tmax (int, optional): tmin and tmax define a window of the signal to compare SNR. Defaults to None.

        Returns:
            dict : A dictionary with the QRF of each component.
        """
        X = signal.comps
        # output = []
        # for Xest in method_output:
        order = order_components(method_output, X)
        Xaux = method_output[order]
        qrfs = []
        for x, xaux in zip(X, Xaux):
            indx = np.where(np.abs(x) > 0)
            qrfs.append(compute_qrf(x[indx], xaux[indx], tmin=tmin, tmax=tmax))
        # output.append(qrfs)
        # output = np.array(output, dtype=object)
        dict_output = {"Comp.{}".format(i): qrfs[i] for i in range(len(qrfs))}
        return dict_output

    @staticmethod
    def compare_instf_block(signal, method_output, tmin=None, tmax=None, **kwargs):
        """ Compute the instantaneous frequency for multi-component outputs.

        Args:
            signal (numpy.array or Signal): _description_
            method_output (numpy.array): _description_
            tmin (int, optional): tmin and tmax define a window of the signal to compare SNR. Defaults to None.
            tmax (int, optional): tmin and tmax define a window of the signal to compare SNR. Defaults to None.

        Returns:
            dict: A dictionary with the IF of each component.
        """
        X = signal.instf
        # output = []
        # for Xest in method_output:
        order = order_components(method_output, X, minormax="min", metric=mse)
        Xaux = method_output[order]
        qrfs = []
        for x, xaux in zip(X, Xaux):
            indx = np.where(np.abs(x) > 0)
            # qrfs.append(compute_qrf(x[indx], xaux[indx],tmin=tmin,tmax=tmax))
            qrfs.append(mse(x[indx], xaux[indx]))
        # output.append(qrfs)
        # output = np.array(output, dtype=object)
        dict_output = {"Comp.{}".format(i): qrfs[i] for i in range(len(qrfs))}
        return dict_output

    @staticmethod
    def sum(bench_a, bench_b):
        """ This function is used to sum to benchmarks by overloading the + operator.
        Summing benchmark means transfer the results of bench_b to bench_a as long as they only differ on the methods/parameters used.

        Args:
            bench_a (Benchmark): The first summand.
            bench_b (Benchmark): The second summand.

        Returns:
            Benchmark: A Benchmark with the combined methods of bench_a and bench_b.
        """
        assert (
            bench_a.repetitions == bench_b.repetitions
        ), "Repetitions should be same in both benchmarks."

        assert (
            bench_a.SNRin == bench_b.SNRin
        ), "SNRin should be the same in both benchmarks."

        assert bench_a.N == bench_b.N, "N should be the same in both benchmarks."

        assert (
            bench_a.Nsub == bench_b.Nsub
        ), "Nsub should be the same in both benchmarks."

        assert np.all(
            bench_a.signal_dic.keys() == bench_b.signal_dic.keys()
        ), "Signals must be the same in both benchmarks."

        assert np.all(
            [key for key in bench_a.objectiveFunction]
            == [key for key in bench_b.objectiveFunction]
        ), "Benchmarks must use the same performance functions."

        bench_c = bench_a

        # Transfer results
        for fun_name in bench_c.objectiveFunction:
            for SNR in bench_c.SNRin:
                for signal_id in bench_c.signal_ids:
                    for method in bench_b.methods:
                        if method not in bench_c.methods.keys():
                            # bench_c.methods[method] = bench_b.methods[method]
                            # bench_c.methods_ids.append(method)
                            # bench_c.parameters[method] = bench_b.parameters[method]
                            # bench_c.this_method_is_new[method] = False
                            bench_c.results[fun_name][signal_id][SNR][method] = {}
                            bench_c.elapsed_time[signal_id][method] = {}
                            for params in bench_b.parameters[method]:
                                bench_c.results[fun_name][signal_id][SNR][method][
                                    str(params)
                                ] = bench_b.results[fun_name][signal_id][SNR][method][
                                    str(params)
                                ]
                                # bench_c.elapsed_time[signal_id][key][str(params)] = bench_b.elapsed_time[signal_id][key][str(params)]

        for method in bench_b.methods:
            if method not in bench_c.methods.keys():
                bench_c.methods[method] = bench_b.methods[method]
                bench_c.methods_ids.append(method)
                bench_c.parameters[method] = bench_b.parameters[method]
                bench_c.this_method_is_new[method] = False
                bench_c.methods_and_params_dic[method] = bench_b.methods_and_params_dic[
                    method
                ]
        return bench_c


""" 
----------------------------------------------------------------------------------------
END OF BENCHMARK CLASS DEFINITION
----------------------------------------------------------------------------------------
"""

"""
----------------------------------------------------------------------------------------
Other auxiliary functions
----------------------------------------------------------------------------------------
"""


def mse(x, xest):
    """ Mean square error performance function.

    Args:
        x (numpy.array): The origninal noiseless signal.
        xest (numpy.array): An estimation of x.

    Returns:
        float: Mean square error between x and xest.
    """
    assert len(x) == len(xest), "Should be of equal length."
    idx = np.where(abs(x) > 0)
    x_aux = x[idx]
    xest_aux = xest[idx]
    error = np.mean((x_aux - xest_aux) ** 2)
    return error


def corr_comps(x, xest):
    """ Normalized correlation between x and xest.

    Args:
        x (numpy.array): The origninal noiseless signal.
        xest (numpy.array): An estimation of x.

    Returns:
        float: Normalized correlation (between -1 and 1).
    """
    idx = np.where(abs(x) > 0)
    x_aux = x[idx]
    xest_aux = xest[idx]
    cor = abs(sum((x_aux - mean(x_aux)) * (xest_aux - mean(xest_aux)))) / (
        norm(x_aux - mean(x_aux)) * norm(xest_aux - mean(xest_aux)) + 1e-15
    )
    return cor


def order_components(Xest, X, minormax="max", metric=corr_comps):
    """ This functions receives a multicomponent output Xest of a method and find a correspondance with the original noiseless components X by minimizing (or maximizing) the metric. 

    Args:
        Xest (numpy.array): _description_
        X (numpy.array): _description_
        minormax (str, optional): 'max' or 'min' according to what is needed from the given metric. Defaults to "max".
        metric (Callable, optional): A function `m = fun(x,xest)`, with m a real number, and vectors x and xest are the noiseless component x and the estimation of x correspondingly. Defaults to correlation between components.

    Returns:
        numpy.array: A vector of lenght K, with the correspondence x[0]<->xest[order[0]].
    """
    order = [[] for aaa in range(len(X))]
    values = np.array([[metric(x, xest) for x in X] for xest in Xest], dtype=object)
    if minormax == "max":
        fun = np.argmax
        factor = -1
    if minormax == "min":
        fun = np.argmin
        factor = 1

    while np.any(np.array([k == [] for k in order], dtype=object)):
        ind = np.unravel_index(fun(values, axis=None), values.shape)
        if (ind[0] not in order) and (order[ind[1]] == []):
            order[ind[1]] = ind[0]
        values[ind] = factor * np.inf
    return order


def compute_qrf(x, x_hat, tmin=None, tmax=None):
    """
    Quality reconstruction factor
    """
    if tmin is None:
        tmin = 0

    if tmax is None:
        tmax = len(x)

    x = x[tmin:tmax]
    x_hat = x_hat[tmin:tmax]
    qrf = 10 * np.log10(np.sum(x**2) / np.sum((x_hat - x) ** 2))
    return qrf


def get_args_and_kwargs(params):
    if type(params) is dict:
        args = []
        kwargs = params
    else:
        dict_indicator = [type(i) is dict for i in params]
        if any(dict_indicator):
            assert (
                len(params) == 2
            ), "Parameters must be given as a dictionary or an iterable."
            for i in range(len(params)):
                kwargs = params[np.where(dict_indicator)[0][0]]
                args = params[np.where([not i for i in dict_indicator])[0][0]]
        else:
            args = params
            kwargs = dict()

    return args, kwargs



"""mcms-benchs: A toolbox that provides a common framework for benchmarks of multi-component signal processing methods.

Copyright (C) 2024  Juan Manuel Miramont

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import importlib.util
import numpy as np
import numbers

# Check matlab.engine is installed
try:
    matlab_is_present = importlib.util.find_spec("matlab")
    if matlab_is_present:
        import matlab.engine

except RuntimeError:
    print("Matlab engine or Matlab installation not found.")


class MatlabInterface:
    """This class offers an interface between python and Matlab to seamlessly run methods in a Benchmark."""

    def __init__(self, matlab_function_name, add2path=[], matlab_warnings=False):
        """Creates a new MatlabInterface method that calls a Matlab function.

        Args:
            matlab_function_name (str): The Matlab function name.
            add2path (list, optional): Add new paths where to look for the function indicated. Defaults to [].
            matlab_warnings (bool, optional): When True, prints out Matlab warnings. Defaults to False.

        Returns:
            MatlabInterface: An object able to call a function implemented in Matlab.
        """
        self.matlab_function_name = matlab_function_name

        try:
            self.eng = matlab.engine.start_matlab()
        except NameError:
            print("Matlab engine or Matlab installation not found.")
            return None

        if not matlab_warnings:
            self.eng.eval("warning('off','all');", nargout=0)

        self.eng.eval("addpath('../src/methods')")
        self.eng.eval("addpath('src/methods')")

        for path in add2path:
            self.eng.eval("addpath('" + path + "')")
            self.eng.eval("addpath(genpath('" + path + "'))")
        # sys.path.insert(0, os.path.abspath('../src/methods/'))

    def matlab_function(self, signal, *params):
        """A wrapper of a Matlab function that receives a signal to process and a variable number of positional arguments.

        Args:
            signal (numpy.ndarray): A numpy array with a signal.

        Returns:
            An equivalent array with the outputs of the Matlab function.
        """
        all_params = list((signal.copy(), *params))
        params = self.pre_parameters(*all_params)
        fun_handler = getattr(self.eng, self.matlab_function_name)
        outputs = fun_handler(*params)
        if isinstance(outputs, numbers.Number):
            return outputs

        if len(outputs) == 1:
            outputs = outputs[0].toarray()
        else:
            outputs = [output.toarray() for output in outputs]
        return np.array(outputs)

    def pre_parameters(self, *params):
        """Cast python types to matlab types before calling the function.

        Returns:
            list: A list of matlab types.
        """
        params_matlab = list()
        for param in params:
            if isinstance(param, np.ndarray):
                params_matlab.append(matlab.double(vector=param.tolist()))
            if isinstance(param, list) or isinstance(param, tuple):
                params_matlab.append(matlab.double(vector=list(param)))
            if isinstance(param, float):
                params_matlab.append(matlab.double(param))
            if isinstance(param, int):
                params_matlab.append(matlab.double(float(param)))

        return params_matlab
